from . import tty
from . import widechar


class KeyValueTable:
    class NewLine:
        pass

    def __init__(self):
        self.rows = []

    def add(self, key, value, color=None, key_color=None, indent=0):
        if value is None:
            lines = []
        elif isinstance(value, (list, tuple)):
            lines = value[:]
        else:
            lines = value.splitlines()

        if not lines:
            lines = [""]

        # add the first line with the key
        self.rows.append((key, lines[0], color, key_color, indent))

        # then add the continuation lines without the key
        for line in lines[1:]:
            self.rows.append(("", line, color, key_color, 0))

    def newline(self):
        self.rows.append((self.NewLine, None, None, None, 0))

    def __str__(self):
        if not self.rows:
            return ""

        col1_width = max([widechar.wc_width(key) + indent for key, _, _, _, indent in self.rows if key != self.NewLine])
        result = []
        skip = False
        for row_num in range(len(self.rows)):
            if skip:
                skip = False
                continue

            key, value, color, key_color, indent = self.rows[row_num]

            if key == self.NewLine:
                result.append("")
                continue

            next_indent = 0  # fake value
            if not value and row_num < len(self.rows) - 1:
                # let's peek if there's a continuation line we could merge instead of the blank value
                next_key, next_value, next_color, next_key_color, next_indent = self.rows[row_num + 1]
                if not next_key:
                    value = next_value
                    color = next_color
                    key_color = next_key_color
                    row_num += 1
                    skip = True

            line = indent * " "

            if not value and next_indent > 0:
                # no value, the key represents a section followed by indented keys -> skip ljust() and " : " separator
                line += tty.colorize(key, key_color)
            else:
                line += tty.colorize(widechar.wc_ljust(key, col1_width - indent), key_color)
                if not key:
                    # continuation line without a key -> skip " : " separator
                    line += "   "
                else:
                    line += " : "
                line += tty.colorize(value, color)

            result.append(line)

        return "\n".join(result)



import osc.commandline


class RequestCommand(osc.commandline.OscCommand):
    """
    Manage requests
    """

    name = "request"
    aliases = ["rq"]

    # arguments specified here will get inherited to all subcommands automatically
    def init_arguments(self):
        self.add_argument(
            "-m",
            "--message",
            metavar="TEXT",
        )



import sys

import osc.commandline_git


class RepoForkCommand(osc.commandline_git.GitObsCommand):
    """
    Fork a git repo
    """

    name = "fork"
    parent = "RepoCommand"

    def init_arguments(self):
        self.add_argument_owner_repo(nargs="+")
        self.add_argument_new_repo_name()

    def run(self, args):
        from osc import gitea_api
        from osc.output import tty

        self.print_gitea_settings()

        if len(args.owner_repo) > 1 and args.new_repo_name:
            self.parser.error("The --new-repo-name option cannot be used with multiple repos")

        num_entries = 0
        failed_entries = []
        for owner, repo in args.owner_repo:
            print(f"Forking git repo {owner}/{repo} ...", file=sys.stderr)
            try:
                response = gitea_api.Fork.create(self.gitea_conn, owner, repo, new_repo_name=args.new_repo_name)
                repo = response.json()
                fork_owner = repo["owner"]["login"]
                fork_repo = repo["name"]
                print(f" * Fork created: {fork_owner}/{fork_repo}", file=sys.stderr)
                num_entries += 1
            except gitea_api.ForkExists as e:
                fork_owner = e.fork_owner
                fork_repo = e.fork_repo
                print(f" * Fork already exists: {fork_owner}/{fork_repo}", file=sys.stderr)
                print(f" * {tty.colorize('WARNING', 'yellow,bold')}: Using an existing fork with a different name than requested", file=sys.stderr)
                num_entries += 1
            except gitea_api.GiteaException as e:
                if e.status == 404:
                    print(f" * {tty.colorize('ERROR', 'red,bold')}: Repo doesn't exist: {owner}/{repo}", file=sys.stderr)
                    failed_entries.append(f"{owner}/{repo}")
                    continue
                raise

        print("", file=sys.stderr)
        print(f"Total forked repos: {num_entries}", file=sys.stderr)
        if failed_entries:
            print(f"{tty.colorize('ERROR', 'red,bold')}: Couldn't fork the following repos: {', '.join(failed_entries)}", file=sys.stderr)
            sys.exit(1)



import osc.commandline


class RequestAcceptCommand(osc.commandline.OscCommand):
    """
    Accept request
    """

    name = "accept"
    parent = "RequestCommand"

    def init_arguments(self):
        self.add_argument(
            "id",
            type=int,
        )

    def run(self, args):
        print(f"Accepting request '{args.id}'")



import osc.commandline_git


class LoginCommand(osc.commandline_git.GitObsCommand):
    """
    Manage configured credentials to Gitea servers
    """

    name = "login"

    def init_arguments(self):
        pass

    def run(self, args):
        self.parser.print_help()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import

from .enums import BlockModes
from .enums import BuildArch
from .enums import LinkedbuildModes
from .enums import RebuildModes
from .repository_download import RepositoryDownload
from .repository_hostsystem import RepositoryHostsystem
from .repository_path import RepositoryPath
from .repository_releasetarget import RepositoryReleasetarget


class Repository(XmlModel):
    XML_TAG = "repository"

    name: str = Field(
        xml_attribute=True,
    )

    rebuild: Optional[RebuildModes] = Field(
        xml_attribute=True,
    )

    block: Optional[BlockModes] = Field(
        xml_attribute=True,
    )

    linkedbuild: Optional[LinkedbuildModes] = Field(
        xml_attribute=True,
    )

    download_list: Optional[List[RepositoryDownload]] = Field(
        xml_name="download",
    )

    releasetarget_list: Optional[List[RepositoryReleasetarget]] = Field(
        xml_name="releasetarget",
    )

    hostsystem_list: Optional[List[RepositoryHostsystem]] = Field(
        xml_name="hostsystem",
    )

    path_list: Optional[List[RepositoryPath]] = Field(
        xml_name="path",
    )

    arch_list: Optional[List[BuildArch]] = Field(
        xml_name="arch",
    )



#!/usr/bin/env python3


"""
This wrapper allows git-obs to be called from the source directory during development.
"""


import osc.commandline_git


if __name__ == "__main__":
    osc.commandline_git.main()



"""
Store class wraps access to files in the '.osc' directory.
It is meant to be used as an implementation detail of Project and Package classes
and shouldn't be used in any code outside osc.
"""


import os
from xml.etree import ElementTree as ET

from . import oscerr
from . import git_scm
from .obs_scm import Store


def get_store(path, check=True, print_warnings=False):
    """
    Return a store object that wraps SCM in given `path`:
     - Store for OBS SCM
     - GitStore for Git SCM
    """

    # if there are '.osc' and '.git' directories next to each other, '.osc' takes preference
    store = None

    try:
        store = Store(path, check)
    except oscerr.NoWorkingCopy:
        pass

    if not store:
        try:
            store = git_scm.GitStore(path, check)
            if print_warnings:
                git_scm.warn_experimental()
        except oscerr.NoWorkingCopy:
            pass

    if not store:
        msg = f"Directory '{path}' is not a working copy"
        raise oscerr.NoWorkingCopy(msg)

    return store



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestSourcediffFileNew(XmlModel):
    XML_TAG = "new"

    name: str = Field(
        xml_attribute=True,
    )

    md5: str = Field(
        xml_attribute=True,
    )

    size: int = Field(
        xml_attribute=True,
    )



import contextlib
import io
import tempfile
import unittest

import osc.conf
from osc.output import KeyValueTable
from osc.output import print_msg
from osc.output import safe_write
from osc.output import sanitize_text
from osc.output import tty


class TestKeyValueTable(unittest.TestCase):
    def test_empty(self):
        t = KeyValueTable()
        self.assertEqual(str(t), "")

    def test_simple(self):
        t = KeyValueTable()
        t.add("Key", "Value")
        t.add("FooBar", "Text")

        expected = """
Key    : Value
FooBar : Text
""".strip()
        self.assertEqual(str(t), expected)

    def test_newline(self):
        t = KeyValueTable()
        t.add("Key", "Value")
        t.newline()
        t.add("FooBar", "Text")

        expected = """
Key    : Value

FooBar : Text
""".strip()
        self.assertEqual(str(t), expected)

    def test_continuation(self):
        t = KeyValueTable()
        t.add("Key", ["Value1", "Value2"])

        expected = """
Key : Value1
      Value2
""".strip()
        self.assertEqual(str(t), expected)

    def test_section(self):
        t = KeyValueTable()
        t.add("Section", None)
        t.add("Key", "Value", indent=4)
        t.add("FooBar", "Text", indent=4)

        expected = """
Section
    Key    : Value
    FooBar : Text
""".strip()
        self.assertEqual(str(t), expected)

    def test_wide_chars(self):
        t = KeyValueTable()
        t.add("Key", "Value")
        t.add("🚀🚀🚀", "Value")

        expected = """
Key    : Value
🚀🚀🚀 : Value
""".strip()
        self.assertEqual(str(t), expected)

    def test_empty_value_no_color(self):
        t = KeyValueTable()
        t.add("Key", "", color="bold")

        expected = "Key : "
        self.assertEqual(str(t), expected)


class TestPrintMsg(unittest.TestCase):
    def setUp(self):
        osc.conf.config = osc.conf.Options()

    def test_debug(self):
        osc.conf.config["debug"] = False
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="debug")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

        osc.conf.config["debug"] = True
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="debug")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("DEBUG: foo bar\n", stderr.getvalue())

    def test_verbose(self):
        osc.conf.config["verbose"] = False
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="verbose")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

        osc.conf.config["verbose"] = True
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="verbose")
        self.assertEqual("foo bar\n", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

        osc.conf.config["verbose"] = False
        osc.conf.config["debug"] = True
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="verbose")
        self.assertEqual("foo bar\n", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

    def test_error(self):
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="error")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual(f"{tty.colorize('ERROR:', 'red,bold')} foo bar\n", stderr.getvalue())

    def test_warning(self):
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="warning")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual(f"{tty.colorize('WARNING:', 'yellow,bold')} foo bar\n", stderr.getvalue())

    def test_none(self):
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to=None)
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

    def test_stdout(self):
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="stdout")
        self.assertEqual("foo bar\n", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())

    def test_stderr(self):
        stdout = io.StringIO()
        stderr = io.StringIO()
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            print_msg("foo", "bar", print_to="stderr")
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("foo bar\n", stderr.getvalue())


class TestSanitization(unittest.TestCase):
    def test_control_chars_bytes(self):
        original = b"".join([i.to_bytes(1, byteorder="big") for i in range(32)])
        sanitized = sanitize_text(original)
        self.assertEqual(sanitized, b"\t\n\r")

    def test_control_chars_str(self):
        original = "".join([chr(i) for i in range(32)])
        sanitized = sanitize_text(original)
        self.assertEqual(sanitized, "\t\n\r")

    def test_csi_escape_sequences_str(self):
        # allowed CSI escape sequences
        originals = [">\033[0m<", ">\033[1;31;47m]<"]
        for original in originals:
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, original)

        # not allowed CSI escape sequences
        originals = [">\033[8m<"]
        for original in originals:
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, "><")

    def test_csi_escape_sequences_bytes(self):
        # allowed CSI escape sequences
        originals = [b">\033[0m<", b">\033[1;31;47m]<"]
        for original in originals:
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, original)

        # not allowed CSI escape sequences
        originals = [b">\033[8m<"]
        for original in originals:
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, b"><")

    def test_standalone_escape_str(self):
        original = ">\033<"
        sanitized = sanitize_text(original)
        self.assertEqual(sanitized, "><")

    def test_standalone_escape_bytes(self):
        # standalone escape
        original = b">\033<"
        sanitized = sanitize_text(original)
        self.assertEqual(sanitized, b"><")

    def test_fe_escape_sequences_str(self):
        for i in range(0x40, 0x5F + 1):
            char = chr(i)
            original = f">\033{char}<"
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, "><")

    def test_fe_escape_sequences_bytes(self):
        for i in range(0x40, 0x5F + 1):
            byte = i.to_bytes(1, byteorder="big")
            original = b">\033" + byte + b"<"
            sanitized = sanitize_text(original)
            self.assertEqual(sanitized, b"><")

    def test_osc_escape_sequences_str(self):
        # OSC (Operating System Command) sequences
        original = "\033]0;this is the window title\007"
        sanitized = sanitize_text(original)
        # \033] is removed with the Fe sequences
        self.assertEqual(sanitized, "0;this is the window title")

    def test_osc_escape_sequences_bytes(self):
        #  OSC (Operating System Command) sequences
        original = b"\033]0;this is the window title\007"
        sanitized = sanitize_text(original)
        # \033] is removed with the Fe sequences
        self.assertEqual(sanitized, b"0;this is the window title")


class TestSafeWrite(unittest.TestCase):
    def test_string_to_file(self):
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            safe_write(f, "string")

    def test_bytes_to_file(self):
        with tempfile.NamedTemporaryFile(mode="wb+") as f:
            safe_write(f, b"bytes")

    def test_string_to_stringio(self):
        with io.StringIO() as f:
            safe_write(f, "string")

    def test_bytes_to_bytesio(self):
        with io.BytesIO() as f:
            safe_write(f, b"bytes")


if __name__ == "__main__":
    unittest.main()



# be careful when debugging this code:
# don't add print statements when setting sys.stdout = SafeWriter(sys.stdout)...
class SafeWriter:
    """
    Safely write an (unicode) str. In case of an "UnicodeEncodeError" the
    the str is encoded with the "encoding" encoding.
    All getattr, setattr calls are passed through to the "writer" instance.
    """

    def __init__(self, writer, encoding='unicode_escape'):
        self._writer = writer
        self._encoding = encoding

    def write(self, s):
        try:
            self._writer.write(s)
        except UnicodeEncodeError as e:
            self._writer.write(s.encode(self._encoding))

    def __getattr__(self, name):
        return getattr(self._writer, name)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)



"""
Store class wraps access to files in the '.osc' directory.
It is meant to be used as an implementation detail of Project and Package classes
and shouldn't be used in any code outside osc.
"""


import os

from .. import oscerr
from .._private import api
from ..util.xml import ET

from typing import List

# __store_version__ is to be incremented when the format of the working copy
# "store" changes in an incompatible way. Please add any needed migration
# functionality to check_store_version().
__store_version__ = '2.0'


class Store:
    STORE_DIR = ".osc"
    STORE_VERSION = "2.0"

    @classmethod
    def is_project_dir(cls, path):
        try:
            store = cls(path)
        except oscerr.NoWorkingCopy:
            return False
        return store.is_project

    @classmethod
    def is_package_dir(cls, path):
        try:
            store = cls(path)
        except oscerr.NoWorkingCopy:
            return False
        return store.is_package

    def __init__(self, path, check=True):
        self.path = path
        self.abspath = os.path.abspath(self.path)

        if check:
            check_store_version(self.abspath)

        self.is_project = self.exists("_project") and not self.exists("_package")
        self.is_package = self.exists("_project") and self.exists("_package")

        if check and not any([self.is_project, self.is_package]):
            msg = f"Directory '{self.path}' is not an OBS SCM working copy"
            raise oscerr.NoWorkingCopy(msg)

    def __contains__(self, fn):
        return self.exists(fn)

    def __iter__(self):
        path = os.path.join(self.abspath, self.STORE_DIR)
        for fn in os.listdir(path):
            full_path = os.path.join(path, fn)
            if os.path.isdir(full_path):
                continue
            yield fn

    def assert_is_project(self):
        if not self.is_project:
            msg = f"Directory '{self.path}' is not an OBS SCM working copy of a project"
            raise oscerr.NoWorkingCopy(msg)

    def assert_is_package(self):
        if not self.is_package:
            msg = f"Directory '{self.path}' is not an OBS SCM working copy of a package"
            raise oscerr.NoWorkingCopy(msg)

    def get_path(self, fn, subdir=None):
        # sanitize input to ensure that joining path works as expected
        fn = fn.lstrip("/")
        if subdir:
            subdir = subdir.lstrip("/")
            return os.path.join(self.abspath, self.STORE_DIR, subdir, fn)
        return os.path.join(self.abspath, self.STORE_DIR, fn)

    def exists(self, fn, subdir=None):
        return os.path.exists(self.get_path(fn, subdir=subdir))

    def unlink(self, fn, subdir=None):
        try:
            os.unlink(self.get_path(fn, subdir=subdir))
        except FileNotFoundError:
            pass

    def read_file(self, fn, subdir=None):
        if not self.exists(fn, subdir=subdir):
            return None
        with open(self.get_path(fn, subdir=subdir), encoding="utf-8") as f:
            return f.read()

    def write_file(self, fn, value, subdir=None):
        if value is None:
            self.unlink(fn, subdir=subdir)
            return
        try:
            if subdir:
                os.makedirs(self.get_path(subdir))
            else:
                os.makedirs(self.get_path(""))
        except FileExistsError:
            pass

        old = self.get_path(fn, subdir=subdir)
        new = self.get_path(f"{fn}.new", subdir=subdir)
        try:
            with open(new, "w", encoding="utf-8") as f:
                f.write(value)
            os.rename(new, old)
        except:
            if os.path.exists(new):
                os.unlink(new)
            raise

    def read_list(self, fn, subdir=None):
        if not self.exists(fn, subdir=subdir):
            return None
        with open(self.get_path(fn, subdir=subdir), encoding="utf-8") as f:
            return [line.rstrip("\n") for line in f]

    def write_list(self, fn, value, subdir=None):
        if value is None:
            self.unlink(fn, subdir=subdir)
            return
        if not isinstance(value, (list, tuple)):
            msg = f"The argument `value` should be list, not {type(value).__name__}"
            raise TypeError(msg)
        value = "".join((f"{line or ''}\n" for line in value))
        self.write_file(fn, value, subdir=subdir)

    def read_string(self, fn, subdir=None):
        if not self.exists(fn, subdir=subdir):
            return None
        with open(self.get_path(fn, subdir=subdir), encoding="utf-8") as f:
            return f.readline().strip()

    def write_string(self, fn, value, subdir=None):
        if value is None:
            self.unlink(fn, subdir=subdir)
            return
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        if not isinstance(value, str):
            msg = f"The argument `value` should be str, not {type(value).__name__}"
            raise TypeError(msg)
        self.write_file(fn, f"{value}\n", subdir=subdir)

    def read_int(self, fn):
        if not self.exists(fn):
            return None
        result = self.read_string(fn)
        if not result.isdigit():
            return None
        return int(result)

    def write_int(self, fn, value, subdir=None):
        if value is None:
            self.unlink(fn, subdir=subdir)
            return
        if not isinstance(value, int):
            msg = f"The argument `value` should be int, not {type(value).__name__}"
            raise TypeError(msg)
        value = str(value)
        self.write_string(fn, value, subdir=subdir)

    def read_xml_node(self, fn, node_name, subdir=None):
        from ..util.xml import xml_parse

        path = self.get_path(fn, subdir=subdir)
        try:
            tree = xml_parse(path)
        except SyntaxError as e:
            msg = f"Unable to parse '{path}': {e}"
            raise oscerr.NoWorkingCopy(msg)
        root = tree.getroot()
        assert root.tag == node_name
        # TODO: return root?
        return tree

    def write_xml_node(self, fn, node_name, node, subdir=None):
        path = self.get_path(fn, subdir=subdir)
        assert node.tag == node_name
        api.write_xml_node_to_file(node, path)

    def _sanitize_apiurl(self, value):
        # apiurl shouldn't end with a slash, strip it so we can use apiurl without modifications
        # in config['api_host_options'][apiurl] and other places
        if isinstance(value, str):
            value = value.strip("/")
        elif isinstance(value, bytes):
            value = value.strip(b"/")
        return value

    @property
    def apiurl(self):
        return self._sanitize_apiurl(self.read_string("_apiurl"))

    @apiurl.setter
    def apiurl(self, value):
        self.write_string("_apiurl", self._sanitize_apiurl(value))

    @property
    def project(self):
        return self.read_string("_project")

    @project.setter
    def project(self, value):
        self.write_string("_project", value)

    @property
    def package(self):
        return self.read_string("_package")

    @package.setter
    def package(self, value):
        self.write_string("_package", value)

    @property
    def scmurl(self):
        return self.read_string("_scm")

    @scmurl.setter
    def scmurl(self, value):
        return self.write_string("_scm", value)

    @property
    def size_limit(self):
        return self.read_int("_size_limit")

    @size_limit.setter
    def size_limit(self, value):
        return self.write_int("_size_limit", value)

    @property
    def to_be_added(self):
        self.assert_is_package()
        return self.read_list("_to_be_added") or []

    @to_be_added.setter
    def to_be_added(self, value):
        self.assert_is_package()
        return self.write_list("_to_be_added", value)

    @property
    def to_be_deleted(self):
        self.assert_is_package()
        return self.read_list("_to_be_deleted") or []

    @to_be_deleted.setter
    def to_be_deleted(self, value):
        self.assert_is_package()
        return self.write_list("_to_be_deleted", value)

    @property
    def in_conflict(self):
        self.assert_is_package()
        return self.read_list("_in_conflict") or []

    @in_conflict.setter
    def in_conflict(self, value):
        self.assert_is_package()
        return self.write_list("_in_conflict", value)

    @property
    def osclib_version(self):
        return self.read_string("_osclib_version")

    @property
    def files(self):
        from .. import core as osc_core

        self.assert_is_package()
        if self.exists("_scm"):
            msg = "Package '{self.path}' is managed via SCM"
            raise oscerr.NoWorkingCopy(msg)
        if not self.exists("_files"):
            msg = "Package '{self.path}' doesn't contain _files metadata"
            raise oscerr.NoWorkingCopy(msg)
        result = []
        directory_node = self.read_xml_node("_files", "directory").getroot()
        for entry_node in api.find_nodes(directory_node, "directory", "entry"):
            result.append(osc_core.File.from_xml_node(entry_node))
        return result

    @files.setter
    def files(self, value):
        if not isinstance(value, (list, tuple)):
            msg = f"The argument `value` should be list, not {type(value).__name__}"
            raise TypeError(msg)

        root = ET.Element("directory")
        for file_obj in sorted(value):
            file_obj.to_xml_node(root)
        self.write_xml_node("_files", "directory", root)

    @property
    def last_buildroot(self):
        self.assert_is_package()
        items = self.read_list("_last_buildroot")
        if items is None:
            return items

        if len(items) != 3:
            msg = f"Package '{self.path}' contains _last_buildroot metadata that doesn't contain 3 lines: [repo, arch, vm_type]"
            raise oscerr.NoWorkingCopy(msg)

        if items[2] in ("", "None"):
            items[2] = None

        return items

    @last_buildroot.setter
    def last_buildroot(self, value):
        self.assert_is_package()
        if len(value) != 3:
            raise ValueError("A list with exactly 3 items is expected: [repo, arch, vm_type]")
        self.write_list("_last_buildroot", value)

    @property
    def _meta_node(self):
        if not self.exists("_meta"):
            return None
        if self.is_package:
            root = self.read_xml_node("_meta", "package").getroot()
        else:
            root = self.read_xml_node("_meta", "project").getroot()
        return root

    def sources_get_path(self, file_name: str) -> str:
        if "/" in file_name:
            raise ValueError(f"Plain file name expected: {file_name}")
        result = os.path.join(self.abspath, self.STORE_DIR, "sources", file_name)
        os.makedirs(os.path.dirname(result), exist_ok=True)
        return result

    def sources_list_files(self) -> List[str]:
        result = []
        invalid = []

        topdir = os.path.join(self.abspath, self.STORE_DIR, "sources")

        if not os.path.isdir(topdir):
            return []

        for fn in os.listdir(topdir):
            if self.sources_is_file(fn):
                result.append(fn)
            else:
                invalid.append(fn)

        if invalid:
            msg = ".osc/sources contains entries other than regular files"
            raise oscerr.WorkingCopyInconsistent(self.project, self.package, invalid, msg)

        return result

    def sources_is_file(self, file_name: str) -> bool:
        return os.path.isfile(self.sources_get_path(file_name))

    def sources_delete_file(self, file_name: str):
        try:
            os.unlink(self.sources_get_path(file_name))
        except:
            pass


store = '.osc'


def check_store_version(dir):
    global store

    versionfile = os.path.join(dir, store, '_osclib_version')
    try:
        with open(versionfile) as f:
            v = f.read().strip()
    except:
        if is_project_dir(dir):
            v = '1.0'
        else:
            v = ''

    if v == '':
        msg = f'Error: "{os.path.abspath(dir)}" is not an osc working copy.'
        if os.path.exists(os.path.join(dir, '.svn')):
            msg = msg + '\nTry svn instead of osc.'
        raise oscerr.NoWorkingCopy(msg)

    if v != __store_version__:
        migrated = False

        if v in ['0.2', '0.3', '0.4', '0.5', '0.6', '0.7', '0.8', '0.9', '0.95', '0.96', '0.97', '0.98', '0.99']:
            # no migration needed, only change metadata version to 1.0
            s = Store(dir, check=False)
            v = "1.0"
            s.write_string("_osclib_version", v)
            migrated = True

        if v == "1.0":
            store_dir = os.path.join(dir, store)
            sources_dir = os.path.join(dir, store, "sources")
            sources_dir_mv = sources_dir

            if os.path.isfile(sources_dir):
                # there is a conflict with an existing "sources" file
                sources_dir_mv = os.path.join(dir, store, "_sources")

            os.makedirs(sources_dir_mv, exist_ok=True)

            s = Store(dir, check=False)
            if s.is_package and not s.scmurl:
                from .package import Package
                from .project import Project

                scm_files = [i.name for i in s.files]

                for fn in os.listdir(store_dir):
                    old_path = os.path.join(store_dir, fn)
                    new_path = os.path.join(sources_dir_mv, fn)
                    if not os.path.isfile(old_path):
                        continue
                    if fn in Package.REQ_STOREFILES or fn in Package.OPT_STOREFILES:
                        continue
                    if fn.startswith("_") and fn not in scm_files:
                        continue
                    if os.path.isfile(old_path):
                        os.rename(old_path, new_path)

            if sources_dir != sources_dir_mv:
                os.rename(sources_dir_mv, sources_dir)

            v = "2.0"
            s.write_string("_osclib_version", v)
            migrated = True

        if migrated:
            return

        msg = f'The osc metadata of your working copy "{dir}"'
        msg += f'\nhas __store_version__ = {v}, but it should be {__store_version__}'
        msg += '\nPlease do a fresh checkout or update your client. Sorry about the inconvenience.'
        raise oscerr.WorkingCopyWrongVersion(msg)


def is_project_dir(d):
    global store

    return os.path.exists(os.path.join(d, store, '_project')) and not \
        os.path.exists(os.path.join(d, store, '_package'))


def is_package_dir(d):
    global store

    return os.path.exists(os.path.join(d, store, '_project')) and \
        os.path.exists(os.path.join(d, store, '_package'))


def read_filemeta(dir):
    from ..util.xml import xml_parse

    global store

    msg = f'\'{dir}\' is not a valid working copy.'
    filesmeta = os.path.join(dir, store, '_files')
    if not is_package_dir(dir):
        raise oscerr.NoWorkingCopy(msg)
    if os.path.isfile(os.path.join(dir, store, '_scm')):
        raise oscerr.NoWorkingCopy("Is managed via scm")
    if not os.path.isfile(filesmeta):
        raise oscerr.NoWorkingCopy(f'{msg} ({filesmeta} does not exist)')

    try:
        r = xml_parse(filesmeta)
    except SyntaxError as e:
        raise oscerr.NoWorkingCopy(f'{msg}\nWhen parsing .osc/_files, the following error was encountered:\n{e}')
    return r


def store_readlist(dir, name):
    global store

    r = []
    if os.path.exists(os.path.join(dir, store, name)):
        with open(os.path.join(dir, store, name)) as f:
            r = [line.rstrip('\n') for line in f]
    return r


def read_tobeadded(dir):
    return store_readlist(dir, '_to_be_added')


def read_tobedeleted(dir):
    return store_readlist(dir, '_to_be_deleted')


def read_sizelimit(dir):
    global store

    r = None
    fname = os.path.join(dir, store, '_size_limit')

    if os.path.exists(fname):
        with open(fname) as f:
            r = f.readline().strip()

    if r is None or not r.isdigit():
        return None
    return int(r)


def read_inconflict(dir):
    return store_readlist(dir, '_in_conflict')


def store_read_project(dir):
    global store

    try:
        with open(os.path.join(dir, store, '_project')) as f:
            p = f.readline().strip()
    except OSError:
        msg = f'Error: \'{os.path.abspath(dir)}\' is not an osc project dir or working copy'
        if os.path.exists(os.path.join(dir, '.svn')):
            msg += '\nTry svn instead of osc.'
        raise oscerr.NoWorkingCopy(msg)
    return p


def store_read_package(dir):
    global store

    try:
        with open(os.path.join(dir, store, '_package')) as f:
            p = f.readline().strip()
    except OSError:
        msg = f'Error: \'{os.path.abspath(dir)}\' is not an osc package working copy'
        if os.path.exists(os.path.join(dir, '.svn')):
            msg += '\nTry svn instead of osc.'
        raise oscerr.NoWorkingCopy(msg)
    return p


def store_read_scmurl(dir):
    import warnings
    warnings.warn(
        "osc.core.store_read_scmurl() is deprecated. "
        "You should be using high-level classes such as Store, Project or Package instead.",
        DeprecationWarning
    )
    return Store(dir).scmurl


def store_read_apiurl(dir, defaulturl=True):
    import warnings
    warnings.warn(
        "osc.core.store_read_apiurl() is deprecated. "
        "You should be using high-level classes such as Store, Project or Package instead.",
        DeprecationWarning
    )
    return Store(dir).apiurl


def store_read_last_buildroot(dir):
    global store

    fname = os.path.join(dir, store, '_last_buildroot')
    if os.path.exists(fname):
        lines = open(fname).read().splitlines()
        if len(lines) == 3:
            return lines

    return


def store_write_string(dir, file, string, subdir=''):
    from ..core import decode_it

    global store

    if subdir and not os.path.isdir(os.path.join(dir, store, subdir)):
        os.mkdir(os.path.join(dir, store, subdir))
    fname = os.path.join(dir, store, subdir, file)
    try:
        f = open(fname + '.new', 'w')
        if not isinstance(string, str):
            string = decode_it(string)
        f.write(string)
        f.close()
        os.rename(fname + '.new', fname)
    except:
        if os.path.exists(fname + '.new'):
            os.unlink(fname + '.new')
        raise


def store_write_project(dir, project):
    store_write_string(dir, '_project', project + '\n')


def store_write_apiurl(dir, apiurl):
    import warnings
    warnings.warn(
        "osc.core.store_write_apiurl() is deprecated. "
        "You should be using high-level classes such as Store, Project or Package instead.",
        DeprecationWarning
    )
    Store(dir).apiurl = apiurl


def store_write_last_buildroot(dir, repo, arch, vm_type):
    store_write_string(dir, '_last_buildroot', repo + '\n' + arch + '\n' + vm_type + '\n')


def store_unlink_file(dir, file):
    global store

    try:
        os.unlink(os.path.join(dir, store, file))
    except:
        pass


def store_read_file(dir, file):
    global store

    try:
        with open(os.path.join(dir, store, file)) as f:
            return f.read()
    except:
        return None


def store_write_initial_packages(dir, project, subelements):
    global store

    fname = os.path.join(dir, store, '_packages')
    root = ET.Element('project', name=project)
    for elem in subelements:
        root.append(elem)
    ET.ElementTree(root).write(fname)


def delete_storedir(store_dir):
    """
    This method deletes a store dir.
    """
    from ..core import delete_dir

    head, tail = os.path.split(store_dir)
    if tail == '.osc':
        delete_dir(store_dir)



import copy
import http.client
import json
import time
import urllib.parse
from typing import Optional

import urllib3
import urllib3.exceptions
import urllib3.response

from .conf import Login


class GiteaHTTPResponse:
    """
    A ``urllib3.response.HTTPResponse`` wrapper
    that ensures compatibility with older versions of urllib3.
    """

    def __init__(self, response: urllib3.response.HTTPResponse):
        self.__dict__["_response"] = response

    def __getattr__(self, name):
        return getattr(self._response, name)

    def json(self):
        if hasattr(self._response, "json"):
            return self._response.json()
        return json.loads(self._response.data)


class Connection:
    def __init__(self, login: Login, alternative_port: Optional[int] = None):
        """
        :param login: ``Login`` object with Gitea url and credentials.
        :param alternative_port: Use an alternative port for the connection. This is needed for testing when gitea runs on a random port.
        """
        self.login = login

        parsed_url = urllib.parse.urlparse(self.login.url, scheme="https")
        if parsed_url.scheme == "http":
            ConnectionClass = urllib3.connection.HTTPConnection
        elif parsed_url.scheme == "https":
            ConnectionClass = urllib3.connection.HTTPSConnection
        else:
            raise ValueError(f"Unsupported scheme in Gitea url '{self.login.url}'")

        self.host = parsed_url.hostname
        assert self.host is not None
        self.port = alternative_port if alternative_port else parsed_url.port
        self.conn = ConnectionClass(host=self.host, port=self.port)

        # retries; variables are named according to urllib3
        self.retry_count = 3
        self.retry_backoff_factor = 2
        self.retry_status_forcelist = (
            500,  # Internal Server Error
            502,  # Bad Gateway
            503,  # Service Unavailable
            504,  # Gateway Timeout
        )

        if hasattr(self.conn, "set_cert"):
            # needed to avoid: AttributeError: 'HTTPSConnection' object has no attribute 'assert_hostname'. Did you mean: 'server_hostname'?
            self.conn.set_cert()

    def makeurl(self, *path: str, query: Optional[dict] = None):
        """
        Return relative url prefixed with "/api/v1/" followed with concatenated ``*path``.
        """
        url_path = ["", "api", "v1"] + [urllib.parse.quote(i, safe="/:") for i in path]
        url_path_str = "/".join(url_path)

        if query is None:
            query = {}
        query = copy.deepcopy(query)

        for key in list(query):
            value = query[key]

            if value in (None, [], ()):
                # remove items with value equal to None or [] or ()
                del query[key]
            elif isinstance(value, bool):
                # convert boolean values to "0" or "1"
                query[key] = str(int(value))

        url_query_str = urllib.parse.urlencode(query, doseq=True)
        return urllib.parse.urlunsplit(("", "", url_path_str, url_query_str, ""))

    def request(self, method, url, json_data: Optional[dict] = None) -> GiteaHTTPResponse:
        """
        Make a request and return ``GiteaHTTPResponse``.
        """
        headers = {
            "Content-Type": "application/json",
        }
        if self.login.token:
            headers["Authorization"] = f"token {self.login.token}"

        if json_data:
            json_data = dict(((key, value) for key, value in json_data.items() if value is not None))

        body = json.dumps(json_data) if json_data else None

        for retry in range(1 + self.retry_count):
            # 1 regular request + ``self.retry_count`` retries
            try:
                self.conn.request(method, url, body, headers)
                response = self.conn.getresponse()

                if response.status not in self.retry_status_forcelist:
                    # we are happy with the response status -> use the response
                    break

                if retry >= self.retry_count:
                    # we have reached maximum number of retries -> use the response
                    break

            except (urllib3.exceptions.HTTPError, ConnectionResetError):
                if retry >= self.retry_count:
                    raise

            # {backoff factor} * (2 ** ({number of previous retries}))
            time.sleep(self.retry_backoff_factor * (2 ** retry))
            self.conn.close()

        if isinstance(response, http.client.HTTPResponse):
            result = GiteaHTTPResponse(urllib3.response.HTTPResponse.from_httplib(response))
        else:
            result = GiteaHTTPResponse(response)

        if not hasattr(response, "status"):
            from .exceptions import GiteaException  # pylint: disable=import-outside-toplevel,cyclic-import

            raise GiteaException(result)

        if response.status // 100 != 2:
            from .exceptions import GiteaException  # pylint: disable=import-outside-toplevel,cyclic-import

            raise GiteaException(result)

        return result



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestSourcediffFileDiff(XmlModel):
    XML_TAG = "diff"

    lines: int = Field(
        xml_attribute=True,
    )

    text: str = Field(
        xml_set_text=True,
    )



import subprocess
import sys

import osc.commandline_git


class RepoCloneCommand(osc.commandline_git.GitObsCommand):
    """
    Clone a git repo

    NOTE: Some of the options may result in setting "core.sshCommand"
    config option in the git repository."
    """

    name = "clone"
    parent = "RepoCommand"

    def init_arguments(self):
        self.add_argument_owner_repo(nargs="+")

        self.add_argument(
            "-a",
            "--anonymous",
            action="store_true",
            default=None,
            help="Clone anonymously via the http protocol",
        )

        self.add_argument(
            "-i",
            "--ssh-key",
            help="Path to a private SSH key (identity file)",
        )

        self.add_argument(
            "--no-ssh-strict-host-key-checking",
            action="store_true",
            help="Set 'StrictHostKeyChecking no' ssh option",
        )

        # TODO: replace with an optional argument to get closer to the `git clone` command?
        self.add_argument(
            "--directory",
            help="Clone into the given directory",
        )

    def run(self, args):
        from osc import gitea_api

        from osc.output import tty

        self.print_gitea_settings()

        if len(args.owner_repo) > 1 and args.directory:
            self.parser.error("The --directory option cannot be used with multiple repos")

        num_entries = 0
        failed_entries = []
        for owner, repo in args.owner_repo:
            print(f"Cloning git repo {owner}/{repo} ...", file=sys.stderr)
            try:
                gitea_api.Repo.clone(
                    self.gitea_conn,
                    owner,
                    repo,
                    directory=args.directory,
                    anonymous=args.anonymous,
                    add_remotes=True,
                    ssh_private_key_path=args.ssh_key or self.gitea_login.ssh_key,
                    ssh_strict_host_key_checking=not(args.no_ssh_strict_host_key_checking),
                )
                num_entries += 1
            except gitea_api.GiteaException as e:
                if e.status == 404:
                    print(f" * {tty.colorize('ERROR', 'red,bold')}: Repo doesn't exist: {owner}/{repo}", file=sys.stderr)
                    failed_entries.append(f"{owner}/{repo}")
                    continue
                raise
            except subprocess.CalledProcessError as e:
                print(f" * {tty.colorize('ERROR', 'red,bold')}: git clone failed", file=sys.stderr)
                failed_entries.append(f"{owner}/{repo}")
                continue

        print("", file=sys.stderr)
        print(f"Total cloned repos: {num_entries}", file=sys.stderr)
        if failed_entries:
            print(f"{tty.colorize('ERROR', 'red,bold')}: Couldn't clone the following repos: {', '.join(failed_entries)}", file=sys.stderr)
            sys.exit(1)



# Copyright 2008,2009 Marcus Huewe <suse-tux@gmx.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation;
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import configparser


# inspired from http://code.google.com/p/iniparse/ - although their implementation is
# quite different

class ConfigLineOrder:
    """
    A ConfigLineOrder() instance task is to preserve the order of a config file.
    It keeps track of all lines (including comments) in the _lines list. This list
    either contains SectionLine() instances or CommentLine() instances.
    """

    def __init__(self):
        self._lines = []

    def _append(self, line_obj):
        self._lines.append(line_obj)

    def _find_section(self, section):
        for line in self._lines:
            if line.type == 'section' and line.name == section:
                return line
        return None

    def add_section(self, sectname):
        self._append(SectionLine(sectname))

    def get_section(self, sectname):
        section = self._find_section(sectname)
        if section:
            return section
        section = SectionLine(sectname)
        self._append(section)
        return section

    def add_other(self, sectname, line):
        if sectname:
            self.get_section(sectname).add_other(line)
        else:
            self._append(CommentLine(line))

    def keys(self):
        return [i.name for i in self._lines if i.type == 'section']

    def __setitem__(self, key, value):
        section = SectionLine(key)
        self._append(section)

    def __getitem__(self, key):
        section = self._find_section(key)
        if not section:
            raise KeyError()
        return section

    def __delitem__(self, key):
        line = self._find_section(key)
        if not line:
            raise KeyError(key)
        self._lines.remove(line)

    def __iter__(self):
        # return self._lines.__iter__()
        for line in self._lines:
            if line.type == 'section':
                yield line.name


class Line:
    """Base class for all line objects"""

    def __init__(self, name, type):
        self.name = name
        self.type = type


class SectionLine(Line):
    """
    This class represents a [section]. It stores all lines which belongs to
    this certain section in the _lines list. The _lines list either contains
    CommentLine() or OptionLine() instances.
    """

    def __init__(self, sectname):
        super().__init__(sectname, 'section')
        self._lines = []

    def _find(self, name):
        for line in self._lines:
            if line.name == name:
                return line
        return None

    def _add_option(self, optname, value=None, line=None, sep='='):
        if value is None and line is None:
            raise configparser.Error('Either value or line must be passed in')
        elif value and line:
            raise configparser.Error('value and line are mutually exclusive')

        if value is not None:
            line = f'{optname}{sep}{value}'
        opt = self._find(optname)
        if opt:
            opt.format(line)
        else:
            self._lines.append(OptionLine(optname, line))

    def add_other(self, line):
        self._lines.append(CommentLine(line))

    def copy(self):
        return dict(self.items())

    def items(self):
        return [(i.name, i.value) for i in self._lines if i.type == 'option']

    def keys(self):
        return [i.name for i in self._lines]

    def __setitem__(self, key, val):
        self._add_option(key, val)

    def __getitem__(self, key):
        line = self._find(key)
        if not line:
            raise KeyError(key)
        return str(line)

    def __delitem__(self, key):
        line = self._find(key)
        if not line:
            raise KeyError(key)
        self._lines.remove(line)

    def __str__(self):
        return self.name

    # XXX: needed to support 'x' in cp._sections['sectname']
    def __iter__(self):
        for line in self._lines:
            yield line.name


class CommentLine(Line):
    """Store a commentline"""

    def __init__(self, line):
        super().__init__(line.strip('\n'), 'comment')

    def __str__(self):
        return self.name


class OptionLine(Line):
    """
    This class represents an option. The class' ``name`` attribute is used
    to store the option's name and the "value" attribute contains the option's
    value. The ``frmt`` attribute preserves the format which was used in the configuration
    file.

    Example::

        optionx:<SPACE><SPACE>value
        => self.frmt = '%s:<SPACE><SPACE>%s'
        optiony<SPACE>=<SPACE>value<SPACE>;<SPACE>some_comment
        => self.frmt = '%s<SPACE>=<SPACE><SPACE>%s<SPACE>;<SPACE>some_comment
    """

    def __init__(self, optname, line):
        super().__init__(optname, 'option')
        self.name = optname
        self.format(line)

    def format(self, line):
        mo = configparser.ConfigParser.OPTCRE.match(line.strip())
        key, val = mo.group('option', 'value')
        self.frmt = line.replace(key.strip(), '%s', 1)
        pos = val.find(' ;')
        if pos >= 0:
            val = val[:pos]
        self.value = val
        self.frmt = self.frmt.replace(val.strip(), '%s', 1).rstrip('\n')

    def __str__(self):
        return self.value


class OscConfigParser(configparser.ConfigParser):
    """
    OscConfigParser() behaves like a normal ConfigParser() object. The
    only differences is that it preserves the order+format of configuration entries
    and that it stores comments.
    In order to keep the order and the format it makes use of the ConfigLineOrder()
    class.
    """

    def __init__(self, defaults=None):
        super().__init__(defaults or {}, interpolation=None)
        self._sections = ConfigLineOrder()

    # XXX: unfortunately we have to override the _read() method from the ConfigParser()
    #      class because a) we need to store comments b) the original version doesn't use
    #      the its set methods to add and set sections, options etc. instead they use a
    #      dictionary (this makes it hard for subclasses to use their own objects, IMHO
    #      a bug) and c) in case of an option we need the complete line to store the format.
    #      This all sounds complicated but it isn't - we only needed some slight changes
    def _read(self, fp, fpname):
        """Parse a sectioned setup file.

        The sections in setup file contains a title line at the top,
        indicated by a name in square brackets (`[]'), plus key/value
        options lines, indicated by `name: value' format lines.
        Continuations are represented by an embedded newline then
        leading whitespace.  Blank lines, lines beginning with a '#',
        and just about everything else are ignored.
        """
        cursect = None                            # None, or a dictionary
        optname = None
        lineno = 0
        e = None                                  # None, or an exception
        while True:
            line = fp.readline()
            if not line:
                break
            lineno = lineno + 1
            # comment or blank line?
            if line.strip() == '' or line[0] in '#;':
                self._sections.add_other(cursect, line)
                continue
            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                # no leading whitespace
                continue
            # continuation line?
            if line[0].isspace() and cursect is not None and optname:
                value = line.strip()
                if value:
                    #cursect[optname] = "%s\n%s" % (cursect[optname], value)
                    #self.set(cursect, optname, "%s\n%s" % (self.get(cursect, optname), value))
                    if cursect == configparser.DEFAULTSECT:
                        self._defaults[optname] = f"{self._defaults[optname]}\n{value}"
                    else:
                        # use the raw value here (original version uses raw=False)
                        self._sections[cursect]._find(optname).value = f'{self.get(cursect, optname, raw=True)}\n{value}'
            # a section header or option header?
            else:
                # is it a section header?
                mo = self.SECTCRE.match(line)
                if mo:
                    sectname = mo.group('header')
                    if self._strict and sectname in self._sections:
                        raise configparser.DuplicateSectionError(sectname, fpname, lineno)
                    elif sectname in self._sections:
                        cursect = self._sections[sectname]
                    elif sectname == configparser.DEFAULTSECT:
                        cursect = self._defaults
                    else:
                        self.add_section(sectname)
                    # So sections can't start with a continuation line
                    cursect = sectname
                    optname = None
                # no section header in the file?
                elif cursect is None:
                    raise configparser.MissingSectionHeaderError(fpname, lineno, line)
                # an option line?
                else:
                    mo = self.OPTCRE.match(line)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        if vi in ('=', ':') and ';' in optval:
                            # ';' is a comment delimiter only if it follows
                            # a spacing character
                            pos = optval.find(';')
                            if pos != -1 and optval[pos - 1].isspace():
                                optval = optval[:pos]
                        optval = optval.strip()
                        # allow empty values
                        if optval == '""':
                            optval = ''
                        optname = self.optionxform(optname.rstrip())
                        if self._strict and optname in self._sections[cursect]:
                            raise configparser.DuplicateOptionError(sectname, optname, fpname, lineno)
                        elif cursect == configparser.DEFAULTSECT:
                            self._defaults[optname] = optval
                        else:
                            self._sections[cursect]._add_option(optname, line=line)
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        if not e:
                            e = configparser.ParsingError(fpname)
                        e.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if e:
            raise e  # pylint: disable-msg=E0702

    def write(self, fp, comments=False):
        """
        write the configuration file. If comments is True all comments etc.
        will be written to fp otherwise the ConfigParsers' default write method
        will be called.
        """
        if comments:
            fp.write(str(self))
            fp.write('\n')
        else:
            super().write(fp)

    def has_option(self, section, option, proper=False, **kwargs):
        """
        Returns True, if the passed section contains the specified option.
        If proper is True, True is only returned if the option is owned by
        this section and not "inherited" from the default.
        """
        if proper:
            return self.optionxform(option) in self._sections[section].keys()
        return super().has_option(section, option, **kwargs)

    # XXX: simplify!
    def __str__(self):
        ret = []
        first = True
        for line in self._sections._lines:
            if line.type == 'section':
                if first:
                    first = False
                else:
                    ret.append('')
                ret.append(f'[{line.name}]')
                for sline in line._lines:
                    if sline.type == 'option':
                        # special handling for continuation lines
                        val = '\n '.join(sline.value.split('\n'))
                        ret.append(sline.frmt % (sline.name, val))
                    elif str(sline) != '':
                        ret.append(str(sline))
            else:
                ret.append(str(line))
        return '\n'.join(ret)

    def _validate_value_types(self, section="", option="", value=""):
        if not isinstance(section, str):
            raise TypeError("section names must be strings")
        if not isinstance(option, str):
            raise TypeError("option keys must be strings")

# vim: sw=4 et



import argparse
import os
import subprocess
import sys

import osc.commandline_common
import osc.commands_git
from . import oscerr
from .output import print_msg


class OwnerRepoAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        from . import gitea_api

        try:
            if isinstance(value, list):
                namespace_value = [gitea_api.Repo.split_id(i) for i in value]
            else:
                namespace_value = gitea_api.Repo.split_id(value)
        except ValueError as e:
            raise argparse.ArgumentError(self, str(e))

        setattr(namespace, self.dest, namespace_value)


class OwnerRepoPullAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        from . import gitea_api

        try:
            if isinstance(value, list):
                namespace_value = [gitea_api.PullRequest.split_id(i) for i in value]
            else:
                namespace_value = gitea_api.PullRequest.split_id(value)
        except ValueError as e:
            raise argparse.ArgumentError(self, str(e))

        setattr(namespace, self.dest, namespace_value)


class BooleanAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        if value is None:
            setattr(namespace, self.dest, None)
        elif value.lower() in ["0", "no", "false", "off"]:
            setattr(namespace, self.dest, False)
        elif value.lower() in ["1", "yes", "true", "on"]:
            setattr(namespace, self.dest, True)
        else:
            raise argparse.ArgumentError(self, f"Invalid boolean value: {value}")


class GitObsCommand(osc.commandline_common.Command):
    @property
    def gitea_conf(self):
        return self.main_command.gitea_conf

    @property
    def gitea_login(self):
        return self.main_command.gitea_login

    @property
    def gitea_conn(self):
        return self.main_command.gitea_conn

    def print_gitea_settings(self):
        print(f"Using the following Gitea settings:", file=sys.stderr)
        print(f" * Config path: {self.gitea_conf.path}", file=sys.stderr)
        print(f" * Login (name of the entry in the config file): {self.gitea_login.name}", file=sys.stderr)
        print(f" * URL: {self.gitea_login.url}", file=sys.stderr)
        print(f" * User: {self.gitea_login.user}", file=sys.stderr)
        print("", file=sys.stderr)

    def add_argument_owner_repo(self, **kwargs):
        self.add_argument(
            "owner_repo",
            action=OwnerRepoAction,
            help="Owner and repo: (format: <owner>/<repo>)",
            **kwargs,
        )

    def add_argument_owner_repo_pull(self, **kwargs):
        self.add_argument(
            "owner_repo_pull",
            action=OwnerRepoPullAction,
            help="Owner, repo and pull request number (format: <owner>/<repo>#<pull-request-number>)",
            **kwargs,
        )

    def add_argument_new_repo_name(self):
        self.add_argument(
            "--new-repo-name",
            help="Name of the newly forked repo",
        )


class GitObsMainCommand(osc.commandline_common.MainCommand):
    name = "git-obs"

    MODULES = (
        ("osc.commands_git", osc.commands_git.__path__[0]),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = None
        self._gitea_conf = None
        self._gitea_login = None
        self._gitea_conn = None

    def init_arguments(self):
        self.add_argument(
            "--gitea-config",
            help="Path to gitea config. Default: $GIT_OBS_CONFIG or ~/.config/tea/config.yml.",
        )

        self.add_argument(
            "-G",
            "--gitea-login",
            help="Name of the login entry in the config file. Default: $GIT_OBS_LOGIN or the default entry from the config file.",
        )

    def post_parse_args(self, args):
        if not args.gitea_config:
            value = os.getenv("GIT_OBS_CONFIG", "").strip()
            if value:
                args.gitea_config = value

        if not args.gitea_login:
            value = os.getenv("GIT_OBS_LOGIN", "").strip()
            if value:
                args.gitea_login = value

        self._args = args


    @classmethod
    def main(cls, argv=None, run=True):
        """
        Initialize OscMainCommand, load all commands and run the selected command.
        """
        cmd = cls()
        cmd.load_commands()
        if run:
            args = cmd.parse_args(args=argv)
            exit_code = cmd.run(args)
            sys.exit(exit_code)
        else:
            args = None
        return cmd, args

    @property
    def gitea_conf(self):
        from . import gitea_api

        if self._gitea_conf is None:
            self._gitea_conf = gitea_api.Config(self._args.gitea_config)
        return self._gitea_conf

    @property
    def gitea_login(self):
        if self._gitea_login is None:
            self._gitea_login = self.gitea_conf.get_login(name=self._args.gitea_login)
        return self._gitea_login

    @property
    def gitea_conn(self):
        from . import gitea_api

        if self._gitea_conn is None:
            self._gitea_conn = gitea_api.Connection(self.gitea_login)
            assert self._gitea_login is not None
        return self._gitea_conn


def main():
    try:
        GitObsMainCommand.main()
    except oscerr.OscBaseError as e:
        print_msg(str(e), print_to="error")
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        print_msg(str(e), print_to="error")
        sys.exit(1)

if __name__ == "__main__":
    main()



import unittest

import osc.conf
from osc.credentials import ObfuscatedConfigFileCredentialsManager


class TestObfuscatedConfigFileCredentialsManager(unittest.TestCase):
    def test_decode_password(self):
        # obfuscated "opensuse"
        password_str = "QlpoOTFBWSZTWeTSblkAAAGBgAIBygAgADDACGNEHxaYXckU4UJDk0m5ZA=="
        password = osc.conf.Password(password_str)
        decoded = ObfuscatedConfigFileCredentialsManager.decode_password(password)
        self.assertEqual(decoded, "opensuse")


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2018 SUSE Linux.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.


import builtins
import html

from .. import oscerr


def decode_list(ilist):
    """ Decodes the elements of a list if needed
    """

    dlist = []
    for elem in ilist:
        if not isinstance(elem, str):
            dlist.append(decode_it(elem))
        else:
            dlist.append(elem)
    return dlist


def decode_it(obj):
    """Decode the given object unless it is a str.

    If the given object is a str or has no decode method, the object itself is
    returned. Otherwise, try to decode the object using utf-8. If this
    fails due to a UnicodeDecodeError, try to decode the object using
    latin-1.
    """
    if isinstance(obj, str) or not hasattr(obj, 'decode'):
        return obj
    try:
        return obj.decode('utf-8')
    except UnicodeDecodeError:
        return obj.decode('latin-1')


def raw_input(*args):
    func = builtins.input

    try:
        return func(*args)
    except EOFError:
        # interpret ctrl-d as user abort
        raise oscerr.UserAbort()


def _html_escape(data):
    return html.escape(data, quote=False)


def format_table(rows, headers):
    """Format list of tuples into equal width table with headers"""
    maxlens = [len(h) for h in headers]
    for r in rows:
        for i, c in enumerate(r):
            maxlens[i] = max(maxlens[i], len(c))
    tpltpl = []
    for i, m in enumerate(maxlens):
        tpltpl.append('{%s:<%s}' % (i, m))
    # {0:12}  {1:7}  {2:10}  {3:8}
    templ = '  '.join(tpltpl) + '\n'

    out = templ.format(*headers)
    out += templ.format(*['-' * m for m in maxlens])
    for r in rows:
        out += templ.format(*r)
    return out



__all__ = [
    'babysitter',
    'build',
    'connection',
    'commandline',
    'core',
    'fetch',
    'grabber',
    'meter',
    'oscerr',
    'oscssl',
]


from .util import git_version
__version__ = git_version.get_version('1.12.1')


# vim: sw=4 et



import osc.commandline_git


class LoginListCommand(osc.commandline_git.GitObsCommand):
    """
    List Gitea credentials entries
    """

    name = "list"
    parent = "LoginCommand"

    def init_arguments(self):
        self.parser.add_argument("--show-tokens", action="store_true", help="Show tokens in the output")

    def run(self, args):
        for login in self.gitea_conf.list_logins():
            print(login.to_human_readable_string(show_token=args.show_tokens))
            print()



import base64
import fcntl
import inspect
import os
import re
import shutil
import subprocess
import ssl
import sys
import tempfile
import time
import warnings

import http.client
import http.cookiejar
import urllib.parse
import urllib.request
import urllib3.exceptions
import urllib3.poolmanager
import urllib3.response
import urllib3.util

from . import __version__
from . import conf
from . import oscerr
from . import oscssl
from . import output
from .util.helper import decode_it


# print only the first occurrence of matching warnings, regardless of location
warnings.filterwarnings("once", category=urllib3.exceptions.InsecureRequestWarning)


class MockRequest:
    """
    Mock a request object for `cookiejar.extract_cookies()`
    and `cookiejar.add_cookie_header()`.
    """

    def __init__(self, url, headers):
        self.url = url
        self.headers = headers
        self.unverifiable = False
        self.type = "https"

    def get_full_url(self):
        return self.url

    def get_header(self, header_name, default=None):
        return self.headers.get(header_name, default)

    def has_header(self, header_name):
        return header_name in self.headers

    def add_unredirected_header(self, key, val):
        # modifies the `headers` variable that was passed to object's constructor
        self.headers[key] = val


def enable_http_debug(config):
    if not int(config["http_debug"]) and not int(config["http_full_debug"]):
        http.client.print = lambda *args, **kwargs: None
        return

    # HACK: override HTTPResponse's init to increase debug level
    old_HTTPResponse__init__ = http.client.HTTPResponse.__init__

    def new_HTTPResponse__init__(self, *args, **kwargs):
        old_HTTPResponse__init__(self, *args, **kwargs)
        self.debuglevel = 1
    http.client.HTTPResponse.__init__ = new_HTTPResponse__init__

    # increase HTTPConnection debug level
    http.client.HTTPConnection.debuglevel = 1

    # HACK: because HTTPResponse's debug data uses print(),
    # let's inject custom print() function to that module
    def new_print(*args, file=None):
        if not int(config["http_full_debug"]) and args:
            # hide private data (authorization and cookies) when full debug is not enabled
            if args[:2] == ("header:", "Set-Cookie:"):
                return
            if args[0] == "send:":
                args = list(args)
                # (?<=...) - '...' must be present before the pattern (positive lookbehind assertion)
                args[1] = re.sub(r"(?<=\\r\\n)authorization:.*?\\r\\n", "", args[1], re.I)
                args[1] = re.sub(r"(?<=\\r\\n)Cookie:.*?\\r\\n", "", args[1], re.I)
        print("DEBUG:", *args, file=sys.stderr)
    http.client.print = new_print


def get_proxy_manager(env):
    proxy_url = os.environ.get(env, None)

    if not proxy_url:
        return

    proxy_purl = urllib3.util.parse_url(proxy_url)

    # rebuild proxy url in order to remove auth because ProxyManager would fail on it
    if proxy_purl.port:
        proxy_url = f"{proxy_purl.scheme}://{proxy_purl.host}:{proxy_purl.port}"
    else:
        proxy_url = f"{proxy_purl.scheme}://{proxy_purl.host}"

    proxy_headers = urllib3.make_headers(
        user_agent=f"osc/{__version__}",
    )

    if proxy_purl.auth:
        proxy_basic_auth = urllib.parse.unquote(proxy_purl.auth)
        proxy_basic_auth = proxy_basic_auth.encode("utf-8")
        proxy_basic_auth = base64.b64encode(proxy_basic_auth).decode()
        proxy_headers["Proxy-Authorization"] = f"Basic {proxy_basic_auth:s}"

    manager = urllib3.ProxyManager(proxy_url, proxy_headers=proxy_headers)
    return manager


# Instantiate on first use in `http_request()`.
# Each `apiurl` requires a differently configured pool
# (incl. trusted keys for example).
CONNECTION_POOLS = {}


# Pool manager for requests outside apiurls.
POOL_MANAGER = urllib3.PoolManager()


# Proxy manager for HTTP connections.
HTTP_PROXY_MANAGER = get_proxy_manager("HTTP_PROXY")


# Proxy manager for HTTPS connections.
HTTPS_PROXY_MANAGER = get_proxy_manager("HTTPS_PROXY")


def http_request_wrap_file(func):
    """
    Turn file path into a file object and close it automatically
    by using a context manager.
    """
    def new_func(method, url, headers=None, data=None, file=None):
        if file:
            with open(file, "rb") as f:
                return func(method, url, headers, data, file=f)
        else:
            return func(method, url, headers, data, file)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    return new_func


@http_request_wrap_file
def http_request(method: str, url: str, headers=None, data=None, file=None):
    """
    Send a HTTP request to a server.

    Features:
    * Authentication ([apiurl]/{user,pass} in oscrc)
    * Session cookie support (~/.local/state/osc/cookiejar)
    * SSL certificate verification incl. managing trusted certs
    * SSL certificate verification bypass (if [apiurl]/sslcertck=0 in oscrc)
    * Expired SSL certificates are no longer accepted. Either prolong them or set sslcertck=0.
    * Proxy support (HTTPS_PROXY env, NO_PROXY is respected)
    * Retries (http_retries in oscrc)
    * Requests outside apiurl (incl. proxy support)
    * Connection debugging (-H/--http-debug, --http-full-debug)

    :param method: HTTP request method (such as GET, POST, PUT, DELETE).
    :param url: The URL to perform the request on.
    :param headers: Dictionary of custom headers to send.
    :param data: Data to send in the request body (conflicts with `file`).
    :param file: Path to a file to send as data in the request body (conflicts with `data`).
    """

    purl = urllib3.util.parse_url(url)
    apiurl = conf.extract_known_apiurl(url)
    headers = urllib3.response.HTTPHeaderDict(headers or {})

    # identify osc
    headers.update(urllib3.make_headers(user_agent=f"osc/{__version__}"))

    if data and file:
        raise RuntimeError('Specify either `data` or `file`')
    elif data:
        if hasattr(data, "encode"):
            data = data.encode("utf-8")
        content_length = len(data)
    elif file:
        content_length = os.fstat(file.fileno()).st_size
        data = file
    else:
        content_length = 0

    if content_length:
        headers.add("Content-Length", str(content_length))

    # handle requests that go outside apiurl
    # do not set auth cookie or auth credentials
    if not apiurl:
        if purl.scheme == "http" and HTTP_PROXY_MANAGER and not urllib.request.proxy_bypass(url):
            # connection through proxy
            manager = HTTP_PROXY_MANAGER
        elif purl.scheme == "https" and HTTPS_PROXY_MANAGER and not urllib.request.proxy_bypass(url):
            # connection through proxy
            manager = HTTPS_PROXY_MANAGER
        else:
            # direct connection
            manager = POOL_MANAGER

        response = manager.urlopen(method, url, body=data, headers=headers, preload_content=False)

        if response.status / 100 != 2:
            raise urllib.error.HTTPError(url, response.status, response.reason, response.headers, response)

        return response

    options = conf.config["api_host_options"][apiurl]

    if options.http_headers:
        new_headers = urllib3.response.HTTPHeaderDict()
        # user-defined headers from the config file
        new_headers.update(options.http_headers)
        # original ``headers`` (Content-Length, User-Agent) must prevail over user-defined headers
        new_headers.update(headers)
        headers = new_headers

    global CONNECTION_POOLS
    pool = CONNECTION_POOLS.get(apiurl, None)
    if not pool:
        pool_kwargs = {}

        # urllib3.Retry() argument 'method_whitelist' got renamed to 'allowed_methods'
        sig = inspect.signature(urllib3.Retry)
        arg_names = list(sig.parameters.keys())
        if "allowed_methods" in arg_names:
            retries_kwargs = {"allowed_methods": None}
        else:
            retries_kwargs = {"method_whitelist": None}

        pool_kwargs["retries"] = urllib3.Retry(
            total=int(conf.config["http_retries"]),
            backoff_factor=2,
            status_forcelist=(
                500,  # Internal Server Error
                502,  # Bad Gateway
                503,  # Service Unavailable
                504,  # Gateway Timeout
            ),
            # don't raise because we want an actual response rather than a MaxRetryError with "too many <status_code> error responses" message
            raise_on_status=False,
            **retries_kwargs,
        )

        if purl.scheme == "https":
            ssl_context = oscssl.create_ssl_context()
            ssl_context.load_default_certs()
            pool_kwargs["ssl_context"] = ssl_context
            # turn cert verification off if sslcertck = 0

            if options["cafile"] or options["capath"]:
                ssl_context.load_verify_locations(cafile=options["cafile"], capath=options["capath"])

            # urllib3 v1
            pool_kwargs["cert_reqs"] = "CERT_REQUIRED" if options["sslcertck"] else "CERT_NONE"

            # urllib3 v2
            if options["sslcertck"]:
                ssl_context.check_hostname = True
                ssl_context.verify_mode = ssl.CERT_REQUIRED
            else:
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE

        if purl.scheme == "http" and HTTP_PROXY_MANAGER and not urllib.request.proxy_bypass(url):
            # connection through HTTP proxy
            pool = HTTP_PROXY_MANAGER.connection_from_host(
                host=purl.host,
                port=purl.port,
                scheme=purl.scheme,
                pool_kwargs=pool_kwargs
            )
            HTTP_PROXY_MANAGER.request('GET', url)
        elif purl.scheme == "https" and HTTPS_PROXY_MANAGER and not urllib.request.proxy_bypass(url):
            # connection through HTTPS proxy
            pool = HTTPS_PROXY_MANAGER.connection_from_host(
                host=purl.host,
                port=purl.port,
                scheme=purl.scheme,
                pool_kwargs=pool_kwargs
            )
        elif purl.scheme == "https":
            # direct connection
            pool = urllib3.HTTPSConnectionPool(host=purl.host, port=purl.port, **pool_kwargs)
        else:
            pool = urllib3.HTTPConnectionPool(host=purl.host, port=purl.port, **pool_kwargs)

        if purl.scheme == "https":
            # inject ssl context instance into pool so we can use it later
            pool.ssl_context = ssl_context

            # inject trusted cert store instance into pool so we can use it later
            pool.trusted_cert_store = oscssl.TrustedCertStore(ssl_context, purl.host, purl.port)

        CONNECTION_POOLS[apiurl] = pool

    auth_handlers = [
        CookieJarAuthHandler(apiurl, os.path.expanduser(conf.config["cookiejar"])),
        SignatureAuthHandler(apiurl, options["user"], options["sshkey"], options["pass"]),
        BasicAuthHandler(apiurl, options["user"], options["pass"]),
    ]

    for handler in auth_handlers:
        # authenticate using a cookie (if available)
        success = handler.set_request_headers(url, headers)
        if success:
            break

    # Rails sends a html response if the header is not set
    # https://github.com/openSUSE/open-build-service/pull/13019
    headers.add("Accept", "application/xml")

    if method == "PUT" or (method == "POST" and (data or file)):
        headers.add("Content-Type", "application/xml; charset=utf-8")
    elif method == "POST":
        headers.add("Content-Type", "application/x-www-form-urlencoded")

    if purl.scheme == "http" and HTTP_PROXY_MANAGER:
        # HTTP proxy requires full URL with 'same host' checking off
        urlopen_url = url
        assert_same_host = False
    else:
        # everything else is fine with path only
        # join path and query, ignore the remaining args; args are (scheme, netloc, path, query, fragment)
        urlopen_url = urllib.parse.urlunsplit(("", "", purl.path, purl.query, ""))
        assert_same_host = True

    if int(conf.config['http_debug']):
        # use the hacked print() for consistency
        http.client.print(40 * '-')
        http.client.print(method, url)

    try:
        response = pool.urlopen(
            method, urlopen_url, body=data, headers=headers,
            preload_content=False, assert_same_host=assert_same_host
        )
    except urllib3.exceptions.MaxRetryError as e:
        if not isinstance(e.reason, urllib3.exceptions.SSLError):
            # re-raise exceptions that are not related to SSL
            raise
        # ssl.SSLCertVerificationError doesn't exist on python 3.6
        # ssl.CertificateError is an alias for ssl.SSLCertVerificationError on python 3.7+
        if isinstance(e.reason.args[0], ssl.CertificateError):
            self_signed_verify_codes = (
                oscssl.X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT,
                oscssl.X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN,
            )
            if e.reason.args[0].verify_code not in self_signed_verify_codes:
                # re-raise ssl exceptions that are not related to self-signed certs
                raise e.reason.args[0] from None
        else:
            # re-raise other than ssl exceptions
            raise e.reason.args[0] from None

        # get the untrusted certificated from server
        cert = pool.trusted_cert_store.get_server_certificate()

        # prompt user if we should trust the certificate
        pool.trusted_cert_store.prompt_trust(cert, reason=e.reason)

        if hasattr(data, 'seek'):
            data.seek(0)

        response = pool.urlopen(
            method, urlopen_url, body=data, headers=headers,
            preload_content=False, assert_same_host=assert_same_host
        )

    if response.status == 401:
        # session cookie has expired, re-authenticate
        for handler in auth_handlers:
            success = handler.set_request_headers_after_401(url, headers, response)
            if success:
                break
        if hasattr(data, 'seek'):
            data.seek(0)
        response = pool.urlopen(method, urlopen_url, body=data, headers=headers, preload_content=False)

    # we want to save a session cookie before an exception is raised on failed requests
    for handler in auth_handlers:
        handler.process_response(url, headers, response)

    if response.status / 100 != 2:
        raise urllib.error.HTTPError(url, response.status, response.reason, response.headers, response)

    return response


# pylint: disable=C0103,C0116
def http_GET(*args, **kwargs):
    return http_request("GET", *args, **kwargs)


# pylint: disable=C0103,C0116
def http_POST(*args, **kwargs):
    return http_request("POST", *args, **kwargs)


# pylint: disable=C0103,C0116
def http_PUT(*args, **kwargs):
    return http_request("PUT", *args, **kwargs)


# pylint: disable=C0103,C0116
def http_DELETE(*args, **kwargs):
    return http_request("DELETE", *args, **kwargs)


class AuthHandlerBase:
    def __init__(self, apiurl):
        self.apiurl = apiurl

    def _get_auth_schemes(self, response):
        """
        Extract all `www-authenticate` headers from `response` and return them
        in a dictionary: `{scheme: auth_method}`.
        """
        result = {}
        for auth_method in response.headers.get_all("www-authenticate", []):
            scheme = auth_method.split()[0].lower()
            result[scheme] = auth_method
        return result

    def set_request_headers(self, url, request_headers):
        """
        Modify request headers with auth headers.

        :param url: Request URL provides context for `request_headers` modifications
        :type  url: str
        :param request_headers: object to be modified
        :type  request_headers: urllib3.response.HTTPHeaderDict
        :return: `True` on if `request_headers` was modified, `False` otherwise
        """
        raise NotImplementedError

    def set_request_headers_after_401(self, url, request_headers, response):
        """
        Modify request headers with auth headers after getting 401 response.

        :param url: Request URL provides context for `request_headers` modifications
        :type  url: str
        :param request_headers: object to be modified
        :type  request_headers: urllib3.response.HTTPHeaderDict
        :param response: Response object provides context for `request_headers` modifications
        :type  response: urllib3.response.HTTPResponse
        :return: `True` on if `request_headers` was modified, `False` otherwise
        """
        raise NotImplementedError

    def process_response(self, url, request_headers, response):
        """
        Retrieve data from response, save cookies, etc.

        :param url: Request URL provides context for `request_headers` modifications
        :type  url: str
        :param request_headers: object to be modified
        :type  request_headers: urllib3.response.HTTPHeaderDict
        :param response: Response object provides context for `request_headers` modifications
        :type  response: urllib3.response.HTTPResponse
        """
        raise NotImplementedError


class CookieJarAuthHandler(AuthHandlerBase):
    # Shared among instances, instantiate on first use, key equals to cookiejar path.
    COOKIEJARS = {}

    def __init__(self, apiurl, cookiejar_path):
        super().__init__(apiurl)
        self.cookiejar_path = cookiejar_path
        if self.cookiejar_path in self.COOKIEJARS:
            self.cookiejar_lock_path = None
        else:
            # Cookiejar hasn't been loaded yet, let's lock it to avoid
            # doing expensive signature auth in multiple processes.
            # This usually happens when a user runs multiple osc instances
            # from the command-line in parallel.
            self.cookiejar_lock_path = f"{self.cookiejar_path}.lock"
        self.cookiejar_lock_fd = None

    @property
    def _cookiejar(self):
        jar = self.COOKIEJARS.get(self.cookiejar_path, None)
        if not jar:
            try:
                os.makedirs(os.path.dirname(self.cookiejar_path), mode=0o700)
            except FileExistsError:
                pass
            jar = http.cookiejar.LWPCookieJar(self.cookiejar_path)
            if os.path.isfile(self.cookiejar_path):
                try:
                    jar.load()
                except http.cookiejar.LoadError:
                    pass
            self.COOKIEJARS[self.cookiejar_path] = jar
        return jar

    def _lock(self):
        if self.cookiejar_lock_path:
            try:
                os.makedirs(os.path.dirname(self.cookiejar_lock_path), mode=0o700)
            except FileExistsError:
                pass
            self.cookiejar_lock_fd = open(self.cookiejar_lock_path, "w")
            fcntl.flock(self.cookiejar_lock_fd, fcntl.LOCK_EX)

    def _unlock(self):
        if self.cookiejar_lock_path:
            self.cookiejar_lock_path = None
            fcntl.flock(self.cookiejar_lock_fd, fcntl.LOCK_UN)
            self.cookiejar_lock_fd.close()

    def set_request_headers(self, url, request_headers):
        self._lock()
        self._cookiejar.add_cookie_header(MockRequest(url, request_headers))
        if request_headers.get_all("cookie", None):
            # we have a valid cookie already -> unlock immediately
            self._unlock()
            return True
        return False

    def set_request_headers_after_401(self, url, request_headers, response):
        # can't do anything, we have tried setting a cookie already
        return False

    def process_response(self, url, request_headers, response):
        if response.headers.get_all("set-cookie", None):
            self._cookiejar.extract_cookies(response, MockRequest(url, response.headers))
            self._cookiejar.save()
        self._unlock()


class BasicAuthHandler(AuthHandlerBase):
    def __init__(self, apiurl, user, password):
        super().__init__(apiurl)
        self.user = user
        self.password = password

    def set_request_headers(self, url, request_headers):
        return False

    def set_request_headers_after_401(self, url, request_headers, response):
        auth_schemes = self._get_auth_schemes(response)
        if "basic" not in auth_schemes:
            return False
        if not self.user or not self.password:
            return False

        basic_auth = f"{self.user:s}:{self.password:s}"
        basic_auth = basic_auth.encode("utf-8")
        basic_auth = base64.b64encode(basic_auth).decode()
        request_headers["Authorization"] = f"Basic {basic_auth:s}"

        return True

    def process_response(self, url, request_headers, response):
        pass


class SignatureAuthHandler(AuthHandlerBase):
    def __init__(self, apiurl, user, sshkey, basic_auth_password=None):
        super().__init__(apiurl)
        self.user = user
        self.sshkey = None
        self.sshkey_fingerprint = None

        if sshkey and re.match("^[A-Z0-9]+:.*", sshkey):
            # if it starts with a prefix such as 'SHA256:' then it's a fingerprint
            self.sshkey_fingerprint = sshkey
        else:
            self.sshkey = sshkey

        if self.sshkey:
            # if only a file name is provided, prepend ~/.ssh
            if "/" not in self.sshkey:
                self.sshkey = os.path.join("~", ".ssh", self.sshkey)
                self.sshkey = os.path.expanduser(self.sshkey)
            output.print_msg(f"Using ssh key file configured in oscrc: {self.sshkey}", print_to="debug")

        self.ssh_keygen_path = shutil.which("ssh-keygen")
        self.ssh_add_path = shutil.which("ssh-add")

        creds_mgr = conf.config["api_host_options"][self.apiurl].get("credentials_mgr_class", None)
        if creds_mgr == "osc.credentials.TransientCredentialsManager":
            self.basic_auth_password = False
        else:
            # value of `basic_auth_password` is only used as a hint if we should skip signature auth
            self.basic_auth_password = bool(basic_auth_password)

        self.temp_pubkey = None

    def list_ssh_agent_keys(self):
        if not self.ssh_add_path:
            return []
        cmd = [self.ssh_add_path, '-L']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        stdout, _ = proc.communicate()
        if proc.returncode == 0 and stdout.strip():
            return stdout.strip().splitlines()
        else:
            return []

    def list_ssh_agent_fingerprints(self):
        if not self.ssh_add_path:
            return []
        cmd = [self.ssh_add_path, '-l']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        stdout, _ = proc.communicate()
        if proc.returncode == 0 and stdout.strip():
            lines = stdout.strip().splitlines()
            return [i.split(" ")[1] for i in lines]
        else:
            return []

    def guess_keyfile(self):
        # `ssh-keygen -Y sign` requires a file with a key which is not available during ssh agent forwarding
        # that's why we need to list ssh-agent's keys and store the first one into a temp file
        keys_in_agent = self.list_ssh_agent_keys()
        if keys_in_agent:
            selected_key = None

            # use ssh key from ssh agent by the specified fingerprint
            if self.sshkey_fingerprint:
                fingerprints_in_agent = self.list_ssh_agent_fingerprints()
                try:
                    indx = fingerprints_in_agent.index(self.sshkey_fingerprint)
                    selected_key = keys_in_agent[indx]
                    output.print_msg(f"Using ssh key from ssh agent that matches fingerprint '{self.sshkey_fingerprint}': {selected_key}", print_to="debug")
                except ValueError:
                    pass

            # use ssh key from ssh agent by key's comment obs=<hostname> matching the hostname of apiurl
            if selected_key is None:
                apiurl_hostname = urllib.parse.urlparse(self.apiurl).hostname
                for key in keys_in_agent:
                    comments = key.strip().split(" ")[2:]
                    pattern = f"obs={apiurl_hostname}"
                    if pattern in comments:
                        selected_key = key
                        output.print_msg(f"Using ssh key from ssh agent that has comment '{pattern}' which matches apiurl '{self.apiurl}': {selected_key}", print_to="debug")
                        break

            # use the first ssh key from ssh agent
            if selected_key is None:
                selected_key = keys_in_agent[0]
                output.print_msg(f"Using the first ssh key from ssh agent (see `ssh-add -L`): {selected_key}", print_to="debug")

            self.temp_pubkey = tempfile.NamedTemporaryFile(mode="w+")
            self.temp_pubkey.write(keys_in_agent[0])
            self.temp_pubkey.flush()
            return self.temp_pubkey.name

        sshdir = os.path.expanduser('~/.ssh')
        keyfiles = ('id_ed25519', 'id_ed25519_sk', 'id_rsa', 'id_ecdsa', 'id_ecdsa_sk', 'id_dsa')
        output.print_msg(f"Searching ssh keys in '{sshdir}' in the following order: {', '.join(keyfiles)}", print_to="debug")
        for keyfile in keyfiles:
            keyfile_path = os.path.join(sshdir, keyfile)
            if os.path.isfile(keyfile_path):
                output.print_msg(f"Using ssh key from file: {keyfile_path}", print_to="debug")
                return keyfile_path
        return None

    def ssh_sign(self, data, namespace, keyfile=None):
        if not keyfile:
            keyfile = self.guess_keyfile()
        if not keyfile:
            raise oscerr.OscIOError(None, "No SSH key configured or auto-detected")
        keyfile = os.path.expanduser(keyfile)
        cmd = [self.ssh_keygen_path, '-Y', 'sign', '-f', keyfile, '-n', namespace, '-q']
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8")
        signature, _ = proc.communicate(data)

        if self.temp_pubkey:
            self.temp_pubkey.close()
            self.temp_pubkey = None

        if proc.returncode:
            raise oscerr.OscIOError(None, 'ssh-keygen signature creation failed: %d' % proc.returncode)

        match = re.match(r"\A-----BEGIN SSH SIGNATURE-----\n(.*)\n-----END SSH SIGNATURE-----", signature, re.S)
        if not match:
            raise oscerr.OscIOError(None, 'could not extract ssh signature')
        return base64.b64decode(match.group(1))

    def get_authorization(self, chal):
        realm = chal.get('realm', '')
        now = int(time.time())
        sigdata = "(created): %d" % now
        signature = self.ssh_sign(sigdata, realm, self.sshkey)
        signature = decode_it(base64.b64encode(signature))
        return 'keyId="%s",algorithm="ssh",headers="(created)",created=%d,signature="%s"' \
            % (self.user, now, signature)

    def add_signature_auth_header(self, req, auth):
        token, challenge = auth.split(' ', 1)
        chal = urllib.request.parse_keqv_list(filter(None, urllib.request.parse_http_list(challenge)))
        auth = self.get_authorization(chal)
        if not auth:
            return False
        auth_val = f'Signature {auth}'
        req.add('Authorization', auth_val)
        return True

    def set_request_headers(self, url, request_headers):
        return False

    def set_request_headers_after_401(self, url, request_headers, response):
        auth_schemes = self._get_auth_schemes(response)

        if "signature" not in auth_schemes:
            # unsupported on server
            return False

        if not self.user:
            return False

        if self.basic_auth_password and "basic" in auth_schemes:
            # prefer basic auth, but only if password is set
            return False

        if not self.ssh_keygen_path:
            output.print_msg("Skipping signature auth because ssh-keygen is not available", print_to="debug")
            return False

        if not self.sshkey_known():
            # ssh key not set, try to guess it
            self.sshkey = self.guess_keyfile()

        if not self.sshkey_known():
            # ssh key cannot be guessed
            return False

        return self.add_signature_auth_header(request_headers, auth_schemes["signature"])

    def process_response(self, url, request_headers, response):
        pass

    def sshkey_known(self):
        return self.sshkey is not None



# Copyright (C) 2008 Novell Inc.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.


import errno
import os
import pdb
import signal
import ssl
import sys
import traceback
from http.client import HTTPException, BadStatusLine
from urllib.error import URLError, HTTPError

import urllib3.exceptions

from . import _private
from . import commandline
from . import conf as osc_conf
from . import core as osc_core
from . import oscerr
from . import output
from .OscConfigParser import configparser
from .oscssl import CertVerificationError
from .util.cpio import CpioError
from .util.helper import decode_it
from .util.packagequery import PackageError

try:
    # import as RPMError because the class "error" is too generic
    # pylint: disable=E0611
    from rpm import error as RPMError
except:
    # if rpm-python isn't installed (we might be on a debian system):
    class RPMError(Exception):
        pass


try:
    from keyring.errors import KeyringLocked
except ImportError:
    # python-keyring is not installed
    class KeyringLocked(Exception):
        pass


# the good things are stolen from Matt Mackall's mercurial


def catchterm(*args):
    raise oscerr.SignalInterrupt


# Signals which should terminate the program safely
for name in 'SIGBREAK', 'SIGHUP', 'SIGTERM':
    num = getattr(signal, name, None)
    if num:
        signal.signal(num, catchterm)


def run(prg, argv=None):
    try:
        try:
            # we haven't parsed options yet, that's why we rely on argv directly
            if "--debugger" in (argv or sys.argv[1:]):
                pdb.set_trace()
            # here we actually run the program
            prg.main(argv)
            return 0
        except:
            # If any of these was set via the command-line options,
            # the config values are expected to be changed accordingly.
            # That's why we're working only with the config.
            if osc_conf.config["traceback"] or osc_conf.config["post_mortem"]:
                traceback.print_exc(file=sys.stderr)
                # we could use http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52215
            # enter the debugger, if desired
            if osc_conf.config["post_mortem"]:
                if sys.stdout.isatty() and not hasattr(sys, 'ps1'):
                    pdb.post_mortem(sys.exc_info()[2])
                else:
                    print('sys.stdout is not a tty. Not jumping into pdb.', file=sys.stderr)
            raise
    except oscerr.SignalInterrupt:
        print('killed!', file=sys.stderr)
    except KeyboardInterrupt:
        print('interrupted!', file=sys.stderr)
        return 130
    except oscerr.UserAbort:
        print('aborted.', file=sys.stderr)
    except oscerr.APIError as e:
        print('BuildService API error:', e.msg, file=sys.stderr)
    except oscerr.LinkExpandError as e:
        print(f'Link "{e.prj}/{e.pac}" cannot be expanded:\n', e.msg, file=sys.stderr)
        print('Use "osc repairlink" to fix merge conflicts.\n', file=sys.stderr)
    except oscerr.WorkingCopyWrongVersion as e:
        print(e, file=sys.stderr)
    except oscerr.NoWorkingCopy as e:
        print(e, file=sys.stderr)
    except HTTPError as e:
        print('Server returned an error:', e, file=sys.stderr)
        if hasattr(e, 'osc_msg'):
            print(e.osc_msg, file=sys.stderr)

        try:
            body = e.read()
        except AttributeError:
            body = ''

        output.print_msg(e.hdrs, print_to="debug")
        output.print_msg(body, print_to="debug")

        if e.code in [400, 403, 404, 500]:
            if b'<summary>' in body:
                msg = body.split(b'<summary>')[1]
                msg = msg.split(b'</summary>')[0]
                msg = _private.api.xml_unescape(msg)
                print(decode_it(msg), file=sys.stderr)
        if e.code >= 500 and e.code <= 599:
            print(f'\nRequest: {e.filename}')
            print('Headers:')
            for h, v in e.hdrs.items():
                if h != 'Set-Cookie':
                    print(f"{h}: {v}")

    except BadStatusLine as e:
        print('Server returned an invalid response:', e, file=sys.stderr)
        print(e.line, file=sys.stderr)
    except HTTPException as e:
        print(e, file=sys.stderr)
    except URLError as e:
        msg = 'Failed to reach a server'
        if hasattr(e, '_osc_host_port'):
            msg += f' ({e._osc_host_port})'
        msg += ':\n'
        print(msg, e.reason, file=sys.stderr)
    except ssl.SSLError as e:
        if 'tlsv1' in str(e):
            print('The python on this system or the server does not support TLSv1.2', file=sys.stderr)
        print("SSL Error:", e, file=sys.stderr)
    except OSError as e:
        # ignore broken pipe
        if e.errno != errno.EPIPE:
            raise
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        print(e, file=sys.stderr)
    except (oscerr.ConfigError, oscerr.NoConfigfile) as e:
        print(e, file=sys.stderr)
    except configparser.Error as e:
        print(e.message, file=sys.stderr)
    except oscerr.OscIOError as e:
        print(e.msg, file=sys.stderr)
        output.print_msg(e.e, print_to="debug")
    except (oscerr.WrongOptions, oscerr.WrongArgs) as e:
        print(e, file=sys.stderr)
        return 2
    except oscerr.ExtRuntimeError as e:
        print(f"{e.file}:", e.msg, file=sys.stderr)
    except oscerr.ServiceRuntimeError as e:
        print(e.msg, file=sys.stderr)
    except oscerr.WorkingCopyOutdated as e:
        print(e, file=sys.stderr)
    except (oscerr.PackageExists, oscerr.PackageMissing, oscerr.WorkingCopyInconsistent) as e:
        print(e.msg, file=sys.stderr)
    except oscerr.PackageInternalError as e:
        print('a package internal error occured\n'
              'please file a bug and attach your current package working copy '
              'and the following traceback to it:', file=sys.stderr)
        print(e.msg, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
    except oscerr.PackageError as e:
        print(str(e), file=sys.stderr)
    except PackageError as e:
        print(f'{e.fname}:', e.msg, file=sys.stderr)
    except RPMError as e:
        print(e, file=sys.stderr)
    except KeyringLocked as e:
        print(e, file=sys.stderr)
    except CertVerificationError as e:
        print(e, file=sys.stderr)
    except urllib3.exceptions.MaxRetryError as e:
        print(e.reason, file=sys.stderr)
    except urllib3.exceptions.ProtocolError as e:
        print(e.args[0], file=sys.stderr)
    except CpioError as e:
        print(e, file=sys.stderr)
    except oscerr.OscBaseError as e:
        print('*** Error:', e, file=sys.stderr)
    if osc_core.MESSAGE_BACKUPS:
        print()
        print("If you lost any edited commit messages due to an error, you may find them here:")
        for path in osc_core.MESSAGE_BACKUPS:
            print(f"  - {path}")
    return 1


def main():
    # avoid buffering output on pipes (bnc#930137) Basically,
    # a "print('foo')" call is translated to a corresponding
    # fwrite call that writes to the stdout stream (cf.
    # string_print (Objects/stringobject.c) and builtin_print
    # (Python/bltinmodule.c)); If no pipe is used, stdout is
    # a tty/refers to a terminal => the stream is line buffered
    # (see _IO_file_doallocate (libio/filedoalloc.c)). If a pipe
    # is used, stdout does not refer to a terminal anymore => the
    # stream is fully buffered by default (see
    # _IO_file_doallocate). The following fdopen call makes
    # stdout line buffered again (at least on systems that
    # support setvbuf - if setvbuf is not supported, the stream
    # remains fully buffered (see PyFile_SetBufSize
    # (Objects/fileobject.c))).
    if not os.isatty(sys.stdout.fileno()):
        sys.stdout = os.fdopen(sys.stdout.fileno(), sys.stdout.mode, 1)
        sys.stderr = os.fdopen(sys.stderr.fileno(), sys.stderr.mode, 1)

    appimage = os.getenv("APPIMAGE", None)
    owd = os.getenv("OWD", None)
    if appimage and owd:
        # OWD stands for Original Working Directory and we need to switch there when running in an AppImage
        # https://docs.appimage.org/packaging-guide/environment-variables.html
        os.chdir(owd)

    sys.exit(run(commandline.OscMainCommand()))

# vim: sw=4 et



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestActionPerson(XmlModel):
    XML_TAG = "person"

    name: str = Field(
        xml_attribute=True,
    )

    role: Optional[str] = Field(
        xml_attribute=True,
    )



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .enums import RequestStates
from .request_review_history import RequestReviewHistory


class RequestReview(XmlModel):
    XML_TAG = "review"

    state: RequestStates = Field(
        xml_attribute=True,
    )

    created: Optional[str] = Field(
        xml_attribute=True,
    )

    by_user: Optional[str] = Field(
        xml_attribute=True,
    )

    by_group: Optional[str] = Field(
        xml_attribute=True,
    )

    by_project: Optional[str] = Field(
        xml_attribute=True,
    )

    by_package: Optional[str] = Field(
        xml_attribute=True,
    )

    who: Optional[str] = Field(
        xml_attribute=True,
    )

    when: Optional[str] = Field(
        xml_attribute=True,
    )

    comment: Optional[str] = Field(
    )

    history_list: Optional[List[RequestReviewHistory]] = Field(
        xml_name="history",
    )

    def get_user_and_type(self):
        if self.by_user:
            return (self.by_user, "user")
        if self.by_group:
            return (self.by_group, "group")
        if self.by_package:
            return (f"{self.by_project}/{self.by_package}", "package")
        if self.by_project:
            return (self.by_project, "project")
        raise RuntimeError("Unable to determine user and its type")



import unicodedata


def wc_width(text):
    result = 0
    for char in text:
        if unicodedata.east_asian_width(char) in ("F", "W"):
            result += 2
        else:
            result += 1
    return result


def wc_ljust(text, width, fillchar=" "):
    text_width = wc_width(text)
    fill_width = wc_width(fillchar)

    while text_width + fill_width <= width:
        text += fillchar
        text_width += fill_width

    return text



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class PackageSourcesFile(XmlModel):
    XML_TAG = "entry"

    name: str = Field(
        xml_attribute=True,
    )

    md5: str = Field(
        xml_attribute=True,
    )

    mtime: int = Field(
        xml_attribute=True,
    )

    size: int = Field(
        xml_attribute=True,
    )

    skipped: Optional[bool] = Field(
        xml_attribute=True,
    )

    def _get_cmp_data(self):
        return (self.name, self.mtime, self.size, self.md5, self.skipped or False)



import sys

from .helper import decode_it


class PackageError(Exception):
    """base class for all package related errors"""

    def __init__(self, fname, msg):
        super().__init__()
        self.fname = fname
        self.msg = msg


class PackageQueries(dict):
    """Dict of package name keys and package query values.  When assigning a
    package query, to a name, the package is evaluated to see if it matches the
    wanted architecture and if it has a greater version than the current value.
    """

    # map debian and rpm arches to common obs arches
    architectureMap = {'i386': ['i586', 'i686'], 'amd64': ['x86_64'], 'ppc64el': ['ppc64le'], 'armv6hl': ['armv6l'], 'armv7hl': ['armv7l']}

    def __init__(self, wanted_architecture):
        self.wanted_architecture = wanted_architecture
        super().__init__()

    def add(self, query):
        """Adds package query to dict if it is of the correct architecture and
        is newer (has a greater version) than the currently assigned package.

        :param query: a PackageQuery
        """
        self.__setitem__(query.name(), query)

    def __setitem__(self, name, query):
        if decode_it(name) != decode_it(query.name()):
            raise ValueError("key '%s' does not match "
                             "package query name '%s'" % (name, query.name()))

        architecture = decode_it(query.arch())

        if (architecture in [self.wanted_architecture, 'noarch', 'all', 'any']
            or self.wanted_architecture in self.architectureMap.get(architecture,
                                                                    [])):
            current_query = self.get(name)

            # if current query does not exist or is older than this new query
            if current_query is None or current_query.vercmp(query) <= 0:
                super().__setitem__(name, query)


class PackageQuery:
    """abstract base class for all package types"""

    def read(self, all_tags=False, *extra_tags):
        """Returns a PackageQueryResult instance"""
        raise NotImplementedError

    # Hmmm... this should be a module function (inherting this stuff
    # does not make much sense) (the same is true for the queryhdrmd5 method)
    @staticmethod
    def query(filename, all_tags=False, extra_rpmtags=(), extra_debtags=(), self_provides=True):
        f = open(filename, 'rb')
        magic = f.read(7)
        f.seek(0)
        extra_tags = ()
        pkgquery = None
        if magic[:4] == b'\xed\xab\xee\xdb':
            from . import rpmquery
            pkgquery = rpmquery.RpmQuery(f)
            extra_tags = extra_rpmtags
        elif magic == b'!<arch>':
            from . import debquery
            pkgquery = debquery.DebQuery(f)
            extra_tags = extra_debtags
        elif magic == b'<update':
            f.close()
            return None
        elif magic[:5] == b'<?xml':
            f.close()
            return None
        # arch tar ball compressed with gz, xz or zst
        elif magic[:5] == b'\375\067zXZ' or magic[:2] == b'\037\213' or magic[:4] == b'(\xb5/\xfd':
            from . import archquery
            pkgquery = archquery.ArchQuery(f)
        else:
            raise PackageError(filename, 'unsupported package type. magic: \'%s\'' % magic)
        pkgqueryresult = pkgquery.read(all_tags, self_provides, *extra_tags)
        f.close()
        return pkgqueryresult

    @staticmethod
    def queryhdrmd5(filename):
        f = open(filename, 'rb')
        magic = f.read(7)
        f.seek(0)
        if magic[:4] == b'\xed\xab\xee\xdb':
            from . import rpmquery
            f.close()
            return rpmquery.RpmQuery.queryhdrmd5(filename)
        return None


class PackageQueryResult:
    """abstract base class that represents the result of a package query"""

    def name(self):
        raise NotImplementedError

    def version(self):
        raise NotImplementedError

    def release(self):
        raise NotImplementedError

    def epoch(self):
        raise NotImplementedError

    def arch(self):
        raise NotImplementedError

    def description(self):
        raise NotImplementedError

    def path(self):
        raise NotImplementedError

    def provides(self):
        raise NotImplementedError

    def requires(self):
        raise NotImplementedError

    def conflicts(self):
        raise NotImplementedError

    def obsoletes(self):
        raise NotImplementedError

    def recommends(self):
        raise NotImplementedError

    def suggests(self):
        raise NotImplementedError

    def supplements(self):
        raise NotImplementedError

    def enhances(self):
        raise NotImplementedError

    def gettag(self, tag):
        raise NotImplementedError

    def vercmp(self, pkgquery):
        raise NotImplementedError

    def canonname(self):
        raise NotImplementedError

    def evr(self):
        evr = self.version()

        if self.release():
            evr += b"-" + self.release()

        epoch = self.epoch()
        if epoch is not None and epoch != 0:
            evr = epoch + b":" + evr
        return evr


def cmp(a, b):
    return (a > b) - (a < b)


if __name__ == '__main__':
    try:
        pkgq = PackageQuery.query(sys.argv[1])
    except PackageError as e:
        print(e.msg)
        sys.exit(2)
    print(pkgq.name())
    print(pkgq.version())
    print(pkgq.release())
    print(pkgq.description())
    print('##########')
    print('\n'.join(pkgq.provides()))
    print('##########')
    print('\n'.join(pkgq.requires()))



import os
import re
import unittest

import osc.core
import osc.oscerr
from osc.util.helper import decode_list

from .common import GET, OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'difffile_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestDiffFiles)


class TestDiffFiles(OscTestCase):
    diff_hdr = 'Index: %s\n==================================================================='

    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def testDiffUnmodified(self):
        """diff an unmodified file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['merge']
        self.__check_diff(p, '', None)

    def testDiffAdded(self):
        """diff an added file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['toadd1']
        exp = """%s
--- toadd1\t(revision 0)
+++ toadd1\t(revision 0)
@@ -0,0 +1,1 @@
+toadd1
""" % (TestDiffFiles.diff_hdr % 'toadd1')
        self.__check_diff(p, exp, None)

    def testDiffRemoved(self):
        """diff a removed file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['somefile']
        exp = """%s
--- somefile\t(revision 2)
+++ somefile\t(working copy)
@@ -1,1 +0,0 @@
-some content
""" % (TestDiffFiles.diff_hdr % 'somefile')
        self.__check_diff(p, exp, None)

    def testDiffMissing(self):
        """diff a missing file (missing files are ignored)"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['missing']
        self.__check_diff(p, '', None)

    def testDiffReplaced(self):
        """diff a replaced file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['replaced']
        exp = """%s
--- replaced\t(revision 2)
+++ replaced\t(working copy)
@@ -1,1 +1,1 @@
-yet another file
+foo replaced
""" % (TestDiffFiles.diff_hdr % 'replaced')
        self.__check_diff(p, exp, None)

    def testDiffSkipped(self):
        """diff a skipped file (skipped files are ignored)"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['skipped']
        self.__check_diff(p, '', None)

    def testDiffConflict(self):
        """diff a file which is in the conflict state"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['foo']
        exp = """%s
--- foo\t(revision 2)
+++ foo\t(working copy)
@@ -1,1 +1,5 @@
+<<<<<<< foo.mine
+This is no test.
+=======
 This is a simple test.
+>>>>>>> foo.r2
""" % (TestDiffFiles.diff_hdr % 'foo')
        self.__check_diff(p, exp, None)

    def testDiffModified(self):
        """diff a modified file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['nochange']
        exp = """%s
--- nochange\t(revision 2)
+++ nochange\t(working copy)
@@ -1,1 +1,2 @@
-This file didn't change.
+This file didn't change but
+is modified.
""" % (TestDiffFiles.diff_hdr % 'nochange')
        self.__check_diff(p, exp, None)

    def testDiffUnversioned(self):
        """diff an unversioned file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['toadd2']
        self.assertRaises(osc.oscerr.OscIOError, self.__check_diff, p, '', None)

    def testDiffAddedMissing(self):
        """diff a file which has satus 'A' but the local file does not exist"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['addedmissing']
        self.assertRaises(osc.oscerr.OscIOError, self.__check_diff, p, '', None)

    def testDiffMultipleFiles(self):
        """diff multiple files"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['nochange', 'somefile']
        exp = """%s
--- nochange\t(revision 2)
+++ nochange\t(working copy)
@@ -1,1 +1,2 @@
-This file didn't change.
+This file didn't change but
+is modified.
%s
--- somefile\t(revision 2)
+++ somefile\t(working copy)
@@ -1,1 +0,0 @@
-some content
""" % (TestDiffFiles.diff_hdr % 'nochange', TestDiffFiles.diff_hdr % 'somefile')
        self.__check_diff(p, exp, None)

    def testDiffReplacedEmptyTodo(self):
        """diff a complete package"""
        self._change_to_pkg('replaced')
        p = osc.core.Package('.')
        exp = """%s
--- replaced\t(revision 2)
+++ replaced\t(working copy)
@@ -1,1 +1,1 @@
-yet another file
+foo replaced
""" % (TestDiffFiles.diff_hdr % 'replaced')
        self.__check_diff(p, exp, None)

    def testDiffBinaryAdded(self):
        """diff an added binary file"""
        self._change_to_pkg('binary')
        p = osc.core.Package('.')
        p.todo = ['binary_added']
        exp = """%s
Binary file 'binary_added' added.
""" % (TestDiffFiles.diff_hdr % 'binary_added')
        self.__check_diff(p, exp, None)

    def testDiffBinaryDeleted(self):
        """diff a deleted binary file"""
        self._change_to_pkg('binary')
        p = osc.core.Package('.')
        p.todo = ['binary_deleted']
        exp = """%s
Binary file 'binary_deleted' deleted.
""" % (TestDiffFiles.diff_hdr % 'binary_deleted')
        self.__check_diff(p, exp, None)

    def testDiffBinaryModified(self):
        """diff a modified binary file"""
        self._change_to_pkg('binary')
        p = osc.core.Package('.')
        p.todo = ['binary']
        exp = """%s
Binary file 'binary' has changed.
""" % (TestDiffFiles.diff_hdr % 'binary')
        self.__check_diff(p, exp, None)

    # diff with revision
    @GET('http://localhost/source/osctest/remote_simple_noadd?rev=3', file='testDiffRemoteNoChange_files')
    def testDiffRemoteNoChange(self):
        """diff against remote revision where no file changed"""
        self._change_to_pkg('remote_simple_noadd')
        p = osc.core.Package('.')
        self.__check_diff(p, '', 3)

    @GET('http://localhost/source/osctest/remote_simple?rev=3', file='testDiffRemoteModified_files')
    @GET('http://localhost/source/osctest/remote_simple/merge?rev=3', file='testDiffRemoteModified_merge')
    def testDiffRemoteModified(self):
        """diff against a remote revision with one modified file"""
        self._change_to_pkg('remote_simple')
        p = osc.core.Package('.')
        exp = """%s
--- merge\t(revision 3)
+++ merge\t(working copy)
@@ -1,3 +1,4 @@
 Is it
 possible to
 merge this file?
+I hope so...
%s
--- toadd1\t(revision 0)
+++ toadd1\t(revision 0)
@@ -0,0 +1,1 @@
+toadd1
""" % (TestDiffFiles.diff_hdr % 'merge', TestDiffFiles.diff_hdr % 'toadd1')
        self.__check_diff(p, exp, 3)

    @GET('http://localhost/source/osctest/remote_simple?rev=3', file='testDiffRemoteDeletedLocalAdded_files')
    def testDiffRemoteNotExistingLocalAdded(self):
        """
        a file which doesn't exist in a remote revision and
        has status A in the wc
        """
        self._change_to_pkg('remote_simple')
        p = osc.core.Package('.')
        exp = """%s
--- toadd1\t(revision 0)
+++ toadd1\t(revision 0)
@@ -0,0 +1,1 @@
+toadd1
""" % (TestDiffFiles.diff_hdr % 'toadd1')
        self.__check_diff(p, exp, 3)

    @GET('http://localhost/source/osctest/remote_simple_noadd?rev=3', file='testDiffRemoteExistingLocalNotExisting_files')
    @GET('http://localhost/source/osctest/remote_simple_noadd/foobar?rev=3', file='testDiffRemoteExistingLocalNotExisting_foobar')
    @GET('http://localhost/source/osctest/remote_simple_noadd/binary?rev=3', file='testDiffRemoteExistingLocalNotExisting_binary')
    def testDiffRemoteExistingLocalNotExisting(self):
        """
        a file doesn't exist in the local wc but exists
        in the remote revision
        """
        self._change_to_pkg('remote_simple_noadd')
        p = osc.core.Package('.')
        exp = """%s
--- foobar\t(revision 3)
+++ foobar\t(working copy)
@@ -1,2 +0,0 @@
-foobar
-barfoo
%s
Binary file 'binary' deleted.
""" % (TestDiffFiles.diff_hdr % 'foobar', TestDiffFiles.diff_hdr % 'binary')
        self.__check_diff(p, exp, 3)

    @GET('http://localhost/source/osctest/remote_localmodified?rev=3', file='testDiffRemoteUnchangedLocalModified_files')
    @GET('http://localhost/source/osctest/remote_localmodified/nochange?rev=3', file='testDiffRemoteUnchangedLocalModified_nochange')
    @GET('http://localhost/source/osctest/remote_localmodified/binary?rev=3', file='testDiffRemoteUnchangedLocalModified_binary')
    def testDiffRemoteUnchangedLocalModified(self):
        """remote revision didn't change, local file is modified"""
        self._change_to_pkg('remote_localmodified')
        p = osc.core.Package('.')
        exp = """%s
--- nochange\t(revision 3)
+++ nochange\t(working copy)
@@ -1,1 +1,2 @@
 This file didn't change.
+oh it does
%s
Binary file 'binary' has changed.
""" % (TestDiffFiles.diff_hdr % 'nochange', TestDiffFiles.diff_hdr % 'binary')
        self.__check_diff(p, exp, 3)

    @GET('http://localhost/source/osctest/remote_simple_noadd?rev=3', file='testDiffRemoteMissingLocalExisting_files')
    def testDiffRemoteMissingLocalExisting(self):
        """
        remote revision misses a file which exists in the local wc (state ' ')"""
        self._change_to_pkg('remote_simple_noadd')
        p = osc.core.Package('.')
        exp = """%s
--- foo\t(revision 0)
+++ foo\t(working copy)
@@ -0,0 +1,1 @@
+This is a simple test.
""" % (TestDiffFiles.diff_hdr % 'foo')
        self.__check_diff(p, exp, 3)

    @GET('http://localhost/source/osctest/remote_localdelete?rev=3', file='testDiffRemoteMissingLocalDeleted_files')
    def testDiffRemoteMissingLocalDeleted(self):
        """
        remote revision misses a file which is marked for
        deletion in the local wc
        """
        # empty diff is expected (svn does the same)
        self._change_to_pkg('remote_localdelete')
        p = osc.core.Package('.')
        self.__check_diff(p, '', 3)

    def __check_diff(self, p, exp, revision=None):
        got = ''
        for i in p.get_diff(revision):
            got += ''.join(decode_list(i))

        # When a hunk header refers to a single line in the "from"
        # file and/or the "to" file, e.g.
        #
        #   @@ -37,37 +41,43 @@
        #   @@ -37,39 +41,41 @@
        #   @@ -37,37 +41,41 @@
        #
        # some systems will avoid repeating the line number:
        #
        #   @@ -37 +41,43 @@
        #   @@ -37,39 +41 @@
        #   @@ -37 +41 @@
        #
        # so we need to canonise the output to avoid false negative
        # test failures.

        # TODO: Package.get_diff should return a consistent format
        #       (regardless of the used python version)
        def __canonise_diff(diff):
            # we cannot use re.M because python 2.6's re.sub does
            # not support a flags argument
            diff = [re.sub(r'^@@ -(\d+) ', '@@ -\\1,\\1 ', line)
                    for line in diff.split('\n')]
            diff = [re.sub(r'^(@@ -\d+,\d+) \+(\d+) ', '\\1 +\\2,\\2 ', line)
                    for line in diff]
            return '\n'.join(diff)

        got = __canonise_diff(got)
        exp = __canonise_diff(exp)
        self.assertEqualMultiline(got, exp)


if __name__ == '__main__':
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class Linkinfo(XmlModel):
    XML_TAG = "linkinfo"

    project: str = Field(
        xml_attribute=True,
    )

    package: str = Field(
        xml_attribute=True,
    )

    lsrcmd5: Optional[str] = Field(
        xml_attribute=True,
    )

    xsrcmd5: Optional[str] = Field(
        xml_attribute=True,
    )

    baserev: Optional[str] = Field(
        xml_attribute=True,
    )

    rev: Optional[str] = Field(
        xml_attribute=True,
    )

    srcmd5: Optional[str] = Field(
        xml_attribute=True,
    )

    error: Optional[str] = Field(
        xml_attribute=True,
    )

    missingok: Optional[bool] = Field(
        xml_attribute=True,
    )
